################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali2rates.py - calculate rates for sequences in a multiple alignment
=====================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

calculate evolutionary rates for sequences in a multiple alignment.

Default output is as a list of pairwise distances. 

REV: is NOT the general 12 parameter model, but - like in PAML -
denotes the 9 parameter GTR model.

Usage
-----

Example::

   python mali2rates.py --help

Type::

   python mali2rates.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import time
import math
import tempfile
import shutil
import CGAT.Experiment as E
import CGAT.Mali as Mali
import CGAT.Genomics as Genomics
import CGAT.WrapperPhylip as WrapperPhylip
import CGAT.WrapperCodeML as WrapperCodeML
import CGAT.RateEstimation as RateEstimation

##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
##-----------------------------------------------------------------------------------
def writeModel( grammar, section, options):
    """write a model to output file."""
    if section in options.write or "all" in options.write:
        outfile = open( options.output_pattern % section, "w" )
        outfile.write( "%s\n" % grammar.getGrammar())
        outfile.close()


def CalculateDistancePOVL( seq1, seq2, gap_chars = (".", "-") ):
    """calculate overlap distance between two sequences."""
    naligned, nunaligned = 0, 0
    for c1, c2 in zip( seq1.upper(), seq2.upper()):
        if c1 in gap_chars or c2 in gap_chars:
            if not (c1 in gap_chars and c2 in gap_chars):
                nunaligned += 1
        else:
            naligned += 1
        
    return 100.0 * float( naligned) / float(nunaligned + naligned), 0

def CalculateDistancePID( seq1, seq2,
                          gap_chars = ("-", "."),
                          mask_chars = ("X"),
                          ):

    nidentical, naligned, nunaligned = 0, 0, 0
    for c1, c2 in zip( seq1.upper(), seq2.upper()):
        if c1 in gap_chars or c2 in gap_chars:
            nunaligned += 1
            continue
        naligned += 1        
        if c1 == c2:
            nidentical += 1

    return 100.0 * float(nidentical) / naligned, 0

def CalculateDistanceJC69( info, do_gamma = False, alpha = None ):
    """return Jukes-Cantor distance.
    """

    p = float(info.mNDifferent) / info.mNAligned
    
    if do_gamma:
        ## not done yet
        distance = 0.75 * alpha * (pow(1 - 4*p/3, -1/ alpha) - 1)
        variance = p*(1 - p)/(pow(1 - 4*p/3, -2/(alpha + 1)) * L)

    else:
        distance = -0.75 * math.log(1.0 - 4.0 * p / 3.0);
        variance = p * ( 1.0 - p) / ( math.pow(1.0 - 4.0 * p / 3, 2.0) * info.mNAligned);
        
    return distance, variance

def CalculateDistanceT92( info ):
    """
    P,Q: transitions, transversions frequencies
    q: G+C content
    
    d = -2q(1 - q)loge(1 - P/[2q(1 - q)] - Q) -[1 -2q(1 -q)]loge(1 - 2Q)/2,(4.18)
    V(d) = [c12P + c32Q - (c1P + c3Q)2]/n,(4.19)
    where c1 = 1/(1 - P/[2q(1 - q)] - Q), c2 = 1/(1 - 2Q), c3 = 2q(1 - q)(c1 - c2) + c2, and q is the G+C content

    Note: result is undefined if
        the number of transversions is >= 0.5
        the G+C content is 0
    
    """
    gc = info.getGCContent()

    # if there are no GC or no AT pairs: result is undefined
    if gc == 0 or gc == 1:
        return -1, -1
    
    wg = 2.0 * gc * (1.0 - gc)

    P = float(info.mNTransitions) / info.mNAligned
    Q = float(info.mNTransversions) / info.mNAligned

    a1 = 1.0 - P / wg - Q;
    if a1 <= 0: return -1, -1
    
    a2 = 1.0 - 2.0 * Q;
    if a2 <= 0: return -1, -1

    # print a1, a2, wg, gc, "p=", P, "q=", Q, str(info)

    distance = -wg * math.log(a1) - 0.5 * (1.0 - wg) * math.log(a2);

    c1 = 1/a1;
    c2 = 1/a2;
    c3 = wg*(c1 - c2) + c2;
    
    variance = (c1 * c1 * P + c3 * c3 * Q - math.pow(c1 * P + c3 * Q, 2.0))/ info.mNAligned;

    return distance, variance

def writePhylipResult( result, options ):
    """write result of phylip run."""
    
    options.stdout.write( "seq1\tseq2\tdist\n" )

    for x in range( len(result.mRowHeaders) - 1 ):
        for y in range( x+1, len(result.mColHeaders) ):
            options.stdout.write( "\t".join( map(str, (
                result.mRowHeaders[x],
                result.mColHeaders[y],
                options.format % result.mMatrix[x,y] ))) + "\n" )

def runDNADIST( mali, pairs, options ):
    """run dnadist."""
    # use phylip for these
    phylip = WrapperPhylip.Phylip()
    phylip.setProgram( "dnadist" )
    phylip.setMali( mali )

    phylip_options = []
    if options.distance == "K80":
        phylip_options += ["D"] * 1
    elif options.distance == "JC69":
        phylip_options += ["D"] * 2
    elif options.distance == "LogDet":
        phylip_options += ["D"] * 3

    phylip_options.append("Y")
    phylip.setOptions( phylip_options )

    if options.dump:
        phylip.setLogLevel(2)
    
    result = phylip.run()

    writePhylipResult( result, options )


def printValue( value, format = "%f" ):
    if value == "na":
        return "na"
    else:
        return format % value

##-------------------------------------------------------------------
##-------------------------------------------------------------------
##-------------------------------------------------------------------
def printPair( pair, mali, map_new2old, options, msg ="" ):
    """print pairs form codeml."""

    ids = mali.getIdentifiers()
    if options.output_format == "list":
        options.stdout.write( "\t".join( (map_new2old[ids[0]],
                                          map_new2old[ids[1]],
                                          options.format % pair.mDistanceMatrix[ids[0]][ids[1]],
                                          options.format % pair.mLogLikelihood,
                                          printValue( pair.mAlpha, options.format),
                                          printValue( pair.mKappa, options.format),
                                          msg ) ) )

    elif options.output_format == "tree":
        options.stdout.write( ">pair%i" % (noutput + 1))
        options.stdout.write( "%s\n" % pair.mTree )

    if options.with_counts:
        info = Genomics.CalculatePairIndices( mali[ids[1]], mali[ids[0]], with_codons = options.is_codons )
        options.stdout.write( "\t%s" % (str(info)) )

    options.stdout.write( "\n" )

    return 1

def runBaseML( mali, pairs, options ):

    baseml = WrapperCodeML.BaseML()

    paml_options = {}

    map_new2old = mali.mapIdentifiers()
    ids = mali.getIdentifiers()

    if options.kappa != None:
        paml_options[ "kappa" ] = str(options.kappa)

    if options.fix_kappa:
        paml_options[ "fix_kappa" ] = "1"

    if options.alpha != None:
       paml_options[ "alpha" ] = str(options.alpha)

    if options.fix_alpha:
        paml_options[ "fix_alpha" ] = "1"

    if options.clean_data:
        paml_options[ "cleandata" ] = options.clean_data

    map_distance2index = {}
    for key,val in baseml.mOptions["model"].items():
        map_distance2index[val] = key
        
    if options.distance.upper() in map_distance2index:
        paml_options[ "model"] = map_distance2index[options.distance]
    else:
        raise "unknown distance for baseml: %s" % options.distance

    if options.filename_tree:
        result = baseml.Run( mali,
                             tree = options.filename_tree,
                             dump = options.dump,
                             test = options.test,
                             options = paml_options )

    elif options.pairwise:
        
        noutput = 0
        ninput = 0
        ntotal = (len(ids) * (len(ids) -1 ) ) / 2

        if options.output_format == "list":
            options.stdout.write( "\t".join( ("seq1", "seq2", "distance", "lnL", "alpha", "kappa", "msg" ) ) )

            if options.with_counts:
                options.stdout.write("\t%s" % Genomics.SequencePairInfo().getHeader() )
            options.stdout.write( "\n" )

        for x,y in pairs:
            m1 = mali.getSequence( ids[x] )            
            ninput += 1
            temp_mali = Mali.Mali()
            m2 = mali.getSequence( ids[y] )

            temp_mali.addSequence( ids[x], m1.mFrom, m1.mTo, m1.mString )
            temp_mali.addSequence( ids[y], m2.mFrom, m2.mTo, m2.mString )

            result = baseml.Run( temp_mali,
                                 tree = "(%s,%s);" % (ids[x], ids[y]),
                                 dump = options.dump,
                                 test = options.test,
                                 options = paml_options )

            if options.loglevel >= 1 and ninput % options.report_step == 0:
                options.stdlog.write( "# pairwise computation: %i/%i -> %i%% in %i seconds.\n" % (ninput, ntotal, 100.0 * ninput / ntotal, time.time() - tstart ) )
                options.stdlog.flush()

            noutput += printPair( result,
                                  temp_mali,
                                  map_new2old,
                                  options )

            options.stdout.flush()
            
    else:
        ## assume that there are only two sequences
        if mali.getLength() == 2:
            id1, id2 = mali.getIdentifiers()
            result = baseml.Run( mali,
                                 tree = "(%s,%s);" % (id1, id2),
                                 dump = options.dump,
                                 test = options.test,
                                 options = paml_options )

        else:
            raise "please supply tree if there are more than two sequences and pairwise mode is not selected."

        if options.output_format == "list":
            all_identifiers = mali.getIdentifiers()
            options.stdout.write( "\t".join( ("seq1", "seq2", "distance", "lnL", "alpha", "kappa" ) ) + "\n" )    
            for x in range( len(all_identifiers) -1 ):
                id_x = all_identifiers[x]
                for y in range(x + 1, len(all_identifiers)):
                    id_y = all_identifiers[y]

                    options.stdout.write( "\t".join( (id_x,
                                                      id_y,
                                                      options.format % result.mDistanceMatrix[id_x][id_y],
                                                      options.format % result.mLogLikelihood,
                                                      options.format % result.mAlpha,
                                                      options.format % result.mKappa ) ) + "\n")

        elif options.output_format == "tree":
            options.stdout.write( "%s\n" % result.mTree )

##---------------------------------------------------------------------------------------
##---------------------------------------------------------------------------------------
##---------------------------------------------------------------------------------------        
def runXrate( mali, pairs, options ):

    from XGram.Generator.Prebuilt import DNA
    from XGram.Model import Annotation
    import XGram.Run

    xgram = XGram.XGram()
    if options.xrate_min_increment:
        xgram.setMinIncrement( options.xrate_min_increment )

    ninput, noutput, nskipped = 0, 0, 0

    tempdir = tempfile.mkdtemp()
    data = tempdir + "/data"

    if options.distance == "K80":
        model = DNA.buildModel( substitution_model = "k80" )
    elif options.distance == "JC69":
        model = DNA.buildModel( substitution_model = "jc69" )        
    elif options.distance == "REV":
        model = DNA.buildModel( substitution_model = "gtr" )        
    else:
        raise "distance %s not implemented for xrate" % (options.distance)

    writeModel( model, "input", options )
    
    if options.output_format == "list":
        options.stdout.write( "\t".join( ("seq1", "seq2", "distance", "lnL", "alpha", "kappa", "msg" ) ) )

        if options.with_counts:
            options.stdout.write("\t%s" % Genomics.SequencePairInfo().getHeader() )
        options.stdout.write( "\n" )

    for x,y in pairs:
        
        m1 = mali.getSequence( ids[x] )            
        ninput += 1
        temp_mali = Mali.Mali()
        m2 = mali.getSequence( ids[y] )

        temp_mali.addSequence( m1.mId, m1.mFrom, m1.mTo, m1.mString )
        temp_mali.addSequence( m2.mId, m2.mFrom, m2.mTo, m2.mString )

##         if temp_mali.getWidth() < options.min_overlap:
##             if options.loglevel >= 1:
##                 options.stdlog.write("# pair %s-%s: not computed because only %i residues overlap\n" % (mali.getEntry(ids[x]).mId,
##                                                                                                         mali.getEntry(ids[y]).mId,
##                                                                                                         temp_mali.getWidth()) )

##             nskipped += 1
##             continue

        outfile = open(data, "w" )
        temp_mali.writeToFile( outfile, format="stockholm",
                               write_ranges = False,
                               options = ("#=GF NH (%s:1.0)%s;" % tuple(temp_mali.getIdentifiers()),) )
        outfile.close()

        o_alpha, o_kappa = "na", "na"
        o_distance = "na"
        msg = ""

        if options.test_xrate:
            for alpha in (0.1, 0.5, 1.0, 1.5):
                for beta in (0.1, 0.5, 1.0, 1.5):
                    model.mGrammar.setParameter( "alpha", alpha )
                    model.mGrammar.setParameter( "beta", beta )                        
                    result = xgram.train( model, data )
                    trained_model = result.getModel()
                    xalpha, xbeta = \
                            ( trained_model.mGrammar.getParameter( 'alpha' ),
                              trained_model.mGrammar.getParameter( 'beta' ) )
                    # this assumes that the branch length in the input is normalized to 1
                    # this is the normalization constant
                    o_distance = options.format % (2 * xbeta + xalpha)
                    o_kappa = options.format % (xalpha / xbeta)

                    msg = "alpha=%6.4f, beta=%6.4f" % (xalpha, xbeta)

                    options.stdout.write( "\t".join( ("%f" % alpha,
                                                      "%f" % beta,
                                                      o_distance,
                                                      options.format % result.getLogLikelihood(),
                                                      o_alpha,
                                                      o_kappa,
                                                      msg) ) )
                    options.stdout.write("\n" )
            continue
        
        options.stdout.write( "%s\t%s\t" % (m1.mId, m2.mId) )
        
        if options.distance in ( "K80", ):
            result = xgram.train( model, data )
            trained_model = result.getModel()

        elif options.distance in ( "REV", ):
            result = xgram.train( model, data )
            trained_model = result.getModel()
            alpha, beta, gamma, delta, epsilon, theta = \
                ( trained_model.mGrammar.getParameter( 'alpha' ),
                  trained_model.mGrammar.getParameter( 'beta' ),
                  trained_model.mGrammar.getParameter( 'gamma' ),
                  trained_model.mGrammar.getParameter( 'delta' ),
                  trained_model.mGrammar.getParameter( 'epsilon' ),
                  trained_model.mGrammar.getParameter( 'theta' ) )

            pi = trained_model.evaluateTerminalFrequencies( ('A0',) )[('A0',)]
            matrix = trained_model.evaluateRateMatrix( ('A0',) )[('A0',)]
            q, d = RateEstimation.getDistanceGTR( pi, matrix )
            o_distance = options.format % (d)
            o_kappa = ""
            msg = "alpha=%6.4f, beta=%6.4f, gamma=%6.4f, delta=%6.4f, epsilon=%6.4f, theta=%6.4f" % (alpha, beta, gamma, delta, epsilon, theta)

        elif options.distance in ('JC69', ):
            result = xgram.buildTree( model, data )

        if options.distance == "K80":
            alpha, beta = \
                   ( trained_model.mGrammar.getParameter( 'alpha' ),
                     trained_model.mGrammar.getParameter( 'beta' ) )
            # this assumes that the branch length in the input is normalized to 1
            # this is the normalization constant
            o_distance = options.format % (2 * beta + alpha)
            o_kappa = options.format % (alpha / beta)

            msg = "alpha=%6.4f, beta=%6.4f" % (alpha, beta)
            alpha = "na"
            
        elif options.distance == "JC69":

            tree = result.getTree()
            ## multiply distance by tree, as rates are set to 1 and 
            ## thus the matrix is scaled by a factor of 3
            o_distance = options.format % (3.0 * float(re.search("\(\S+:([0-9.]+)\)", tree).groups()[0]))
            o_kappa = "na"
            msg = ""

        writeModel( result.mModel, "trained", options )
    
        options.stdout.write("\t".join( (o_distance,
                                         options.format % result.getLogLikelihood(),
                                         o_alpha,
                                         o_kappa,
                                         msg) ) )

        if options.with_counts:
            info = Genomics.CalculatePairIndices( mali[ids[x]], mali[ids[y]], with_codons = options.is_codons )
            options.stdout.write( "\t%s" % (str(info)) )

        options.stdout.write( "\n" )
    
    shutil.rmtree( tempdir )

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: mali2rates.py 2781 2009-09-10 11:33:14Z andreas $",
                                    usage = globals()["__doc__"])

    parser.add_option("-i", "--input-format", dest="input_format", type="choice",
                      choices=("plain", "fasta", "clustal", "stockholm", "phylip" ),
                      help="input format of multiple alignment"  )

    parser.add_option("-s", "--sites", dest="sites", type="string",
                      help="sites to use [default=%default].", )

    parser.add_option("-f", "--file", dest="filename", type="string",
                      help="filename of multiple alignment (- for stdin) [default=%default].",
                      metavar = "FILE" )

    parser.add_option("-o", "--format", dest="format", type="string",
                      help="format [default=%default].",
                      metavar = "format" )

    parser.add_option("-d", "--distance", dest="distance", type="choice",
                      choices=("PID", "T92", "JC69", "POVL", "F84", "LogDet",
                               "K80", "F81", "HKY85", "TN93", "REV", "UNREST", "REVU",
                               "UNRESTU",
                               "JTT", "PMB", "PAM", "Kimura", "CategoriesModel" ),
                      help="method to use for distance calculation [default=%default]." )

    parser.add_option( "--method", dest="method", type="choice",
                      choices=("phylip", "baseml", "own", "xrate"),
                      help = "program to use for rate calculation.")

    parser.add_option( "--output-format", dest="output_format", type="choice",
                      choices=("list", "tree"),
                      help = "output format.")
                     
    parser.add_option("-m", "--min-sites", dest="min_sites", type="int",
                      help="minimum number of sites for output[default=%default].", )

    parser.add_option("-a", "--alphabet", dest="alphabet", type="choice",
                      choices=("aa", "na", "auto"),
                      help="alphabet to use.", )

    parser.add_option("-t", "--filename-tree", dest="filename_tree", type="string",
                      help="filename with tree information."  )

    parser.add_option( "--set-alpha", dest="alpha", type="float",
                      help="initial alpha value.")

    parser.add_option( "--fix-alpha", dest="fix_alpha", action = "store_true",
                       help="do not estimate alpha.")

    parser.add_option( "--set-kappa", dest="kappa", type="float",
                       help="initial kappa value.")

    parser.add_option( "--fix-kappa", dest="fix_kappa", action = "store_true",
                       help="do not estimate kappa.")

    parser.add_option( "--dump", dest="dump", action="store_true",
                      help="dump output."  )

    parser.add_option( "--test", dest="test", action="store_true",
                      help="test run - does not clean up."  )

    parser.add_option( "--pairwise", dest="pairwise", action="store_true",
                      help="force pairwise comparison.")

    parser.add_option( "--set-clean-data", dest="clean_data", type="choice",
                       choices = ("0", "1"),
                       help="PAML should cleanup data:  0=only gaps within pair are removed, 1=columns in the mali with gaps are removed.")

    parser.add_option( "--with-counts", dest="with_counts", action="store_true",
                      help="output counts of aligned positions, transitions and transversions."  )

    parser.add_option("-w", "--write", dest="write", type="choice", action="append",
                      choices=("input", "trained", "all" ),
                      help="output sections to write for xrate."  )
    
    parser.add_option( "--output-pattern", dest="output_pattern", type="string",
                      help="output pattern for output files." )

    parser.add_option( "--xrate-min-increment", dest="xrate_min_increment", type = float,
                       help="minimum increment to stop iteration in xrate." )

    parser.set_defaults( \
        input_format = "fasta",
        filename_tree = None,
        with_counts = False,
        sites = "d4",
        distance = "T92",
        min_sites = 1,
        filename = "-",
        alphabet="auto",
        format= "%6.4f",
        method="phylip",
        kappa = None,
        fix_kappa = False,
        alpha = None,
        fix_alpha = False,
        dump = False,
        clean_data = None,
        output_format = "list",
        iteration="all-vs-all",
        pairwise=False,
        report_step = 1000,
        output_pattern = "%s.eg",
        write = [],
        test_xrate = False,
        xrate_min_increment = None,
        is_codons = False,
        )
    
    (options, args) = E.Start( parser )

    if options.filename != "-":
        infile = open(options.filename, "r")
    else:
        infile = sys.stdin

    ## read multiple alignment
    if options.pairwise:
        ## read sequences, but not as a multiple alignment. This permits multiple names.
        mali = Mali.SequenceCollection()
        options.iteration = "pairwise"
    else:
        mali = Mali.Mali()

    mali.readFromFile( infile, format = options.input_format )
    
    ids = mali.getIdentifiers()

    if options.alphabet == "auto":
        s = "".join( map( lambda x: x.mString, mali.values())).lower()
        ss = re.sub( "[acgtxn]", "", s)
        if float(len(ss)) < (len(s) * 0.1):
            options.alphabet = "na"
            if mali.getNumColumns() % 3 == 0:
                options.is_codons = True
        else:
            options.alphabet = "aa"

        if options.loglevel >= 1:
            options.stdlog.write("# autodetected alphabet: %s\n" % options.alphabet)
            
    if options.filename != "-":
        infile.close()

    npairs = 0
    nskipped_length = 0
    nskipped_distance = 0

    pairs = []
    if options.iteration == "all-vs-all":
        for x in range(len(ids) - 1):
            for y in range(x+1, len(ids)):
                pairs.append( (x,y) )
    elif options.iteration == "first-vs-all":
        for y in range(1, len(ids)):
            pairs.append( (0,y) )
    elif options.iteration == "pairwise":
        if len(ids) % 2 != 0:
            raise "uneven number of sequences (%i) not compatible with --iteration=pairwise" % len(ids)
        for x in range(0, len(ids), 2):
            pairs.append( (x, x+1) )

    if options.alphabet == "na":

        if options.method=="baseml":
            runBaseML( mali, pairs, options )
        elif options.method == "phylip" and options.distance in ("F84", "K80", "JC69", "LogDet" ):
            runDNADIST( mali, pairs, options )
        elif options.method == "xrate":
            runXrate( mali, pairs, options )
        else:
            if options.is_codons:
                h = Genomics.SequencePairInfoCodons().getHeader()
            else:
                h = Genomics.SequencePairInfo().getHeader()
            options.stdout.write( "seq1\tseq2\tdist\tvar\t%s\n" % ( h ))

            for x, y in pairs:
                id_x = ids[x]
                npairs += 1

                id_y = ids[y]

                info = Genomics.CalculatePairIndices( mali[id_x], mali[id_y], with_codons = options.is_codons )

                if options.distance in ("T92", "JC69"):
                    if options.sites == "d4":
                        seq1, seq2 = Genomics.GetDegenerateSites( mali[id_x], mali[id_y], position=3, degeneracy=4 )

                        if len(seq1) < options.min_sites:
                            nskipped_length += 1
                            continue
                    else:
                        raise "unknown sites %s" % options.sites

                if options.distance == "T92":
                    distance, variance = CalculateDistanceT92( info )
                elif options.distance == "JC69":
                    distance, variance = CalculateDistanceJC69( info )
                elif options.distance == "PID":
                    distance, variance = CalculateDistancePID( mali[id_x], mali[id_y] )
                elif options.distance == "POVL":
                    distance, variance = CalculateDistancePOVL( mali[id_x], mali[id_y] )

                if distance >= 0:
                    options.stdout.write( "\t".join( map(str, (id_x, id_y,
                                                               options.format % distance,
                                                               options.format % variance, info) ) ) + "\n" )
                else:
                    nskipped_distance += 1

    elif options.alphabet == "aa":

        if options.distance in ( "JTT", "PMB", "PAM", "Kimura", "CategoriesModel" ):

            # use phylip for these
            phylip = WrapperPhylip.Phylip()
            phylip.setProgram( "protdist" )
            phylip.setMali( mali )
            
            phylip_options = []
            if options.distance == "PMG":
                phylip_options += ["D"] * 1
            elif options.distance == "PAM":
                phylip_options += ["D"] * 2
            elif options.distance == "Kimura":
                phylip_options += ["D"] * 3
            elif options.distance == "CategoriesModel":
                phylip_options += ["D"] * 4
                
            phylip_options.append("Y")
            phylip.setOptions( phylip_options )
            result = phylip.run()

            writePhylipResult( result, options )

        else:
            options.stdout.write( "id1\tid2\tdist\tvar\n" )

            ## iterate over all pairs of sequences
            for x, y in pairs:
                id_x = ids[x]
                npairs += 1

                id_y = ids[y]

                if options.distance == "PID":
                    distance, variance = CalculateDistancePID( mali[id_x], mali[id_y] )
                elif options.distance == "POVL":
                    ## percentage overlap
                    distance, variance = CalculateDistancePOVL( mali[id_x], mali[id_y] )

                if distance >= 0:
                    options.stdout.write("\t".join( (id_x, id_y,
                                      options.format % distance,
                                      options.format % variance) ) + "\n" ) 
                else:
                    nskipped_distance += 1
        
    if options.loglevel >= 1:
        options.stdlog.write("# nseqs=%i, npairs=%i, nskipped_length=%i, nskipped_distance=%i\n" % ( len(ids), npairs, nskipped_length, nskipped_distance))
    
    E.Stop()

"""
/* Copyright 2005 Emmanuel Paradis <paradis@isem.univ-montp2.fr> */

/* This file is part of the `ape' library for R and related languages. */
/* It is made available under the terms of the GNU General Public */
/* License, version 2, or at your option, any later version, */
/* incorporated herein by reference. */

/* This program is distributed in the hope that it will be */
/* useful, but WITHOUT ANY WARRANTY; without even the implied */
/* warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR */
/* PURPOSE.  See the GNU General Public License for more */
/* details. */

/* You should have received a copy of the GNU General Public */
/* License along with this program; if not, write to the Free */
/* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, */
/* MA 02111-1307, USA */

#include <R.h>

#define both_non_n(a, b) (strcmp(a, "n") && strcmp(b, "n"))

void dist_dna_JC69(char **x, int *n, int *s, double *d, int *pairdel,
		   int *variance, double *var, int *gamma, double *alpha)
{
  int i, j, k, s1, s2, target, Nd, L;
  double p;

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) Nd += 1;
      }
      if (!*pairdel) L = *s;
      p = ((double) Nd/L);
      if (*gamma) d[target] = 0.75 * *alpha * (pow(1 - 4*p/3, -1/ *alpha) - 1);
      else d[target] = -0.75 * log(1 - 4 * p/3);
      if (*variance) {
	if (*gamma) var[target] = p*(1 - p)/(pow(1 - 4*p/3, -2/(*alpha + 1)) * L);
	else var[target] = p*(1 - p)/(pow(1 - 4*p/3, 2)*L);
      }
    }
  }
}

void dist_dna_K80(char **x, int *n, int *s, double *d, int *pairdel,
		   int *variance, double *var, int *gamma, double *alpha)
{
  int i, j, k, s1, s2, target, Nd, Ns, L;
  double P, Q, a1, a2, b, c1, c2, c3;

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Ns = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a") && !strcmp(x[s2], "g"))
	    Ns += 1;
	  else {
	    if (!strcmp(x[s1], "g") && !strcmp(x[s2], "a"))
	      Ns += 1;
	    else {
	      if (!strcmp(x[s1], "c") && !strcmp(x[s2], "t"))
		Ns += 1;
	      else {
		if (!strcmp(x[s1], "t") && !strcmp(x[s2], "c"))
		  Ns += 1;
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L = *s;
      P = ((double) Ns/L);
      Q = ((double) (Nd - Ns)/L);
      a1 = 1 - 2 * P - Q;
      a2 = 1 - 2 * Q;
      if (*gamma) {
	b = -1 / *alpha;
	d[target] = *alpha * (pow(a1, b) + 0.5 * pow(a2, b) - 1.5) / 2;
      }
      else d[target] = -0.5 * log(a1 * sqrt(a2));
      if (*variance) {
	if (*gamma) {
	  b = -(1 / *alpha + 1);
	  c1 = pow(a1, b);
	  c2 = pow(a2, b);
	  c3 = (c1 + c2) / 2;
	} else {
	  c1 = 1 / a1;
	  c2 = 1 / a2;
	  c3 = (c1 + c2) / 2;
	}
	var[target] = (c1*c1*P + c3*c3*Q - pow(c1*P + c3*Q, 2))/L;
      }
    }
  }
}

void dist_dna_F81(char **x, int *n, int *s, double *d, double *BF,
		  int *pairdel, int *variance, double *var,
		  int *gamma, double *alpha)
{
  int i, j, k, Nd, L, target, s1, s2;
  double E, p;

  E = 1 - BF[0]*BF[0] - BF[1]*BF[1] - BF[2]*BF[2] - BF[3]*BF[3];

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) Nd += 1;
      }
      if (!*pairdel) L = *s;
      p = ((double) Nd/L);
      if (*gamma) d[target] = E * *alpha * (pow(1 - p/E, -1/ *alpha) - 1);
      else d[target] = -E*log(1 - p/E);
      if (*variance) {
	if (*gamma) var[target] = p*(1 - p)/(pow(1 - p/E, -2/(*alpha + 1)) * L);
	else var[target] = p*(1 - p)/(pow(1 - p/E, 2)*L);
      }
    }
  }
}

void dist_dna_K81(char **x, int *n, int *s, double *d, int *pairdel,
		   int *variance, double *var)
{
  int i, j, k, Nd, Nv1, Nv2, L, s1, s2, target;
  double P, Q, R, a1, a2, a3, a, b, c;

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Nv1 = 0;
      Nv2 = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a")) {
	    if (!strcmp(x[s2], "t")) Nv1 += 1;
	    if (!strcmp(x[s2], "c")) Nv2 += 1;
	  } else {
	    if (!strcmp(x[s1], "g")) {
	      if (!strcmp(x[s2], "c")) Nv1 += 1;
	      if (!strcmp(x[s2], "t")) Nv2 += 1;
	    } else {
	      if (!strcmp(x[s1], "c")) {
		if (!strcmp(x[j * *s + k], "g")) Nv1 += 1;
		if (!strcmp(x[j * *s + k], "a")) Nv2 += 1;
	      } else {
		if (!strcmp(x[s1], "t")) {
		  if (!strcmp(x[s2], "a")) Nv1 += 1;
		  if (!strcmp(x[s2], "g")) Nv2 += 1;
		}
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L = *s;
      P = ((double) (Nd - Nv1 - Nv2)/L);
      Q = ((double) Nv1/L);
      R = ((double) Nv2/L);
      a1 = 1 - 2*P - 2*Q;
      a2 = 1 - 2*P - 2*R;
      a3 = 1 - 2*Q - 2*R;
      d[target] = -0.25 * log(a1 * a2 * a3);
      if (*variance) {
        a = (1/a1 + 1/a2)/2;
	b = (1/a1 + 1/a3)/2;
	c = (1/a2 + 1/a3)/2;
        var[target] = (a*a*P + b*b*Q + c*c*R - pow(a*P + b*Q + c*R, 2))/2;
      }
    }
  }
}

void dist_dna_F84(char **x, int *n, int *s, double *d, double *BF,
		  int *pairdel, int *variance, double *var)
{
  int i, j, k, Nd, Ns, L, target, s1, s2;
  double P, Q, A, B, C, a, b, t1, t2, t3;

  A = (BF[0] * BF[2])/(BF[0] + BF[2]) + (BF[1] * BF[3])/(BF[1] + BF[3]);
  B = BF[0] * BF[2] + BF[1] * BF[3];
  C = (BF[0] + BF[2]) * (BF[1] + BF[3]);

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Ns = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a") && !strcmp(x[s2], "g"))
	    Ns += 1;
	  else {
	    if (!strcmp(x[s1], "g") && !strcmp(x[s2], "a"))
	      Ns += 1;
	    else {
	      if (!strcmp(x[s1], "c") && !strcmp(x[s2], "t"))
		Ns += 1;
	      else {
		if (!strcmp(x[s1], "t") && !strcmp(x[s2], "c"))
		  Ns += 1;
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L = *s;
      P = ((double) Ns/L);
      Q = ((double) (Nd - Ns)/L);
      d[target] = -2*A*log(1 - (P/(2*A) - (A - B)*Q/(2*A*C))) + 2*(A - B - C)*log(1 - Q/(2*C));
      if (*variance) {
        t1 = A*C;
	t2 = C*P/2;
	t3 = (A - B)*Q/2;
        a = t1/(t1 - t2 - t3);
	b = A*(A - B)/(t1 - t2 - t3) - (A - B - C)/(C - Q/2);
	var[target] = (a*a*P + b*b*Q - pow(a*P + b*Q, 2))/2;
      }
    }
  }
}

void dist_dna_T92(char **x, int *n, int *s, double *d, double *BF,
                  int *pairdel, int *variance, double *var)
{
  int i, j, k, Nd, Ns, L, target, s1, s2;
  double P, Q, wg, a1, a2, c1, c2, c3;

  wg = 2 * (BF[1] + BF[2]) * (1 - (BF[1] + BF[2]));

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Ns = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a") && !strcmp(x[s2], "g"))
	    Ns += 1;
	  else {
	    if (!strcmp(x[s1], "g") && !strcmp(x[s2], "a"))
	      Ns += 1;
	    else {
	      if (!strcmp(x[s1], "c") && !strcmp(x[s2], "t"))
		Ns += 1;
	      else {
		if (!strcmp(x[s1], "t") && !strcmp(x[s2], "c"))
		  Ns += 1;
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L = *s;
      P = ((double) Ns/L);
      Q = ((double) (Nd - Ns)/L);
      a1 = 1 - P/wg - Q;
      a2 = 1 - 2*Q;
      d[target] = -wg*log(a1) - 0.5*(1 - wg)*log(a2);
      if (*variance) {
        c1 = 1/a1;
        c2 = 1/a2;
        c3 = wg*(c1 - c2) + c2;
        var[target] = (c1*c1*P + c3*c3*Q - pow(c1*P + c3*Q, 2))/L;
      }
    }
  }
}

void dist_dna_TN93(char **x, int *n, int *s, double *d, double *BF,
		  int *pairdel, int *variance, double *var,
		  int *gamma, double *alpha)
{
  int i, j, k, Nd, Ns1, Ns2, L, target, s1, s2;
  double P1, P2, Q, A, B, C, gR, gY, k1, k2, k3, k4, w1, w2, w3, c1, c2, c3, c4, b;

  gR = BF[0] + BF[2];
  gY = BF[1] + BF[3];
  k1 = 2 * BF[0] * BF[2] / gR;
  k2 = 2 * BF[1] * BF[3] / gY;
  k3 = 2 * (gR * gY - BF[0]*BF[2]*gY/gR - BF[1]*BF[3]*gR/gY);

  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Ns1 = 0;
      Ns2 = 0;
      L = 0;
      for (k = 0; k < *s; k++) {
        s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a") && !strcmp(x[s2], "g"))
	    Ns1 += 1;
	  else {
	    if (!strcmp(x[s1], "g") && !strcmp(x[s2], "a"))
	      Ns1 += 1;
	    else {
	      if (!strcmp(x[s1], "c") && !strcmp(x[s2], "t"))
		Ns2 += 1;
	      else {
		if (!strcmp(x[s1], "t") && !strcmp(x[s2], "c"))
		  Ns2 += 1;
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L = *s;
      P1 = ((double) Ns1/L);
      P2 = ((double) Ns2/L);
      Q = ((double) (Nd - Ns1 - Ns2)/L);
      w1 = 1 - P1/k1 - Q/(2*gR);
      w2 = 1 - P2/k2 - Q/(2*gY);
      w3 = 1 - Q/(2*gR*gY);
      if (*gamma) {
        k4 = 2*(BF[0]*BF[2] + BF[1]*BF[3] + gR*gY);
        b = -1 / *alpha;
        c1 = pow(w1, b);
        c2 = pow(w2, b);
        c3 = pow(w3, b);
        c4 = k1*c1/(2*gR) + k2*c2/(2*gY) + k3*c3/(2*gR*gY);
        d[target] = *alpha * (k1*pow(w1, b) + k2*pow(w2, b) + k3*pow(w3, b) - k4);
      } else {
        k4 = 2*((BF[0]*BF[0] + BF[2]*BF[2])/(2*gR*gR) + (BF[2]*BF[2] + BF[3]*BF[3])/(2*gY*gY));
        c1 = 1/w1;
        c2 = 1/w2;
        c3 = 1/w3;
        c4 = k1 * c1/(2 * gR) + k2 * c2/(2 * gY) + k4 * c3;
        d[target] = -k1*log(w1) - k2*log(w2) - k3*log(w3);
      }
      if (*variance) {
        var[target] = (c1*c1*P1 + c2*c2*P2 + c4*c4*Q - pow(c1*P1 + c2*P2 + c4*Q, 2))/L;
      }
    }
  }
}

void dist_dna_GG95(char **x, int *n, int *s, double *d,
		   int *pairdel, int *variance, double *var)
{
  int i, j, k, s1, s2, target, *GC, gccount, Nd, Ns, *L, length, tl, npair;
  double *theta, gcprop, *P, pp, *Q, qq, *tstvr, svr, A,
    sum, ma /* mean alpha */, K1, K2;

  GC = &gccount;
  theta = &gcprop;
  L = &length;
  P = &pp;
  Q = &qq;
  tstvr = &svr;

  npair = *n * (*n - 1) / 2;

  GC = (int*)R_alloc(*n, sizeof(int));
  theta = (double*)R_alloc(*n, sizeof(double));
  L = (int*)R_alloc(npair, sizeof(int));
  P = (double*)R_alloc(npair, sizeof(double));
  Q = (double*)R_alloc(npair, sizeof(double));
  tstvr = (double*)R_alloc(npair, sizeof(double));

  /* get the proportion of GC, theta, in each sequence */
  for (i = 0; i < *n; i++) {
    tl = 0;
    GC[i] = 0;
    for (k = 0; k < *s; k++) {
      s1 = i * *s + k;
      if (*pairdel) {
	if (strcmp(x[s1], "n")) tl += 1;
	else continue;
      }
      if (!strcmp(x[s1], "c") || !strcmp(x[s1], "g"))
	GC[i] += 1;
    }
    if (!*pairdel) tl = *s;
    theta[i] = ((double) GC[i]/tl);
  }

  /* get the proportions of transitions and transversions,
     and the estimates of their ratio for each pair; we
     also get the sample size for each pair in L  */
  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      Nd = 0;
      Ns = 0;
      L[target] = 0;
      for (k = 0; k < *s; k++) {
	s1 = i * *s + k;
	s2 = j * *s + k;
	if (*pairdel) {
	  if (both_non_n(x[s1], x[s2])) L[target] += 1;
	  else continue;
	}
	if (strcmp(x[s1], x[s2])) {
	  Nd += 1;
	  if (!strcmp(x[s1], "a") && !strcmp(x[s2], "g"))
	    Ns += 1;
	  else {
	    if (!strcmp(x[s1], "g") && !strcmp(x[s2], "a"))
	      Ns += 1;
	    else {
	      if (!strcmp(x[s1], "c") && !strcmp(x[s2], "t"))
		Ns += 1;
	      else {
		if (!strcmp(x[s1], "t") && !strcmp(x[s2], "c"))
		  Ns += 1;
	      }
	    }
	  }
	}
      }
      if (!*pairdel) L[target] = *s;
      P[target] = ((double) Ns/L[target]);
      Q[target] = ((double) (Nd - Ns)/L[target]);
      A = log(1 - 2*Q[target]);
      tstvr[target] = 2*(log(1 - 2*P[target] - Q[target]) - 0.5*A)/A;
    }
  }

  /* compute the mean alpha (ma) = mean Ts/Tv */
  sum = 0;
  tl = 0;
  for (i = 0; i < npair; i++)
    /* some values of tstvr are -inf if there is no
       transversions observed: we exclude them */
    if (R_FINITE(tstvr[i])) {
      sum += tstvr[i];
      tl += 1;
    }
  ma = sum/tl;

  /* compute the distance for each pair */
  for (i = 0; i < *n - 1; i++) {
    for (j = i + 1; j < *n; j++) {
      target = *n * i - i*(i + 1)/2 + j - i - 1;
      A = 1 - 2*Q[target];
      K1 = 1 + ma*(theta[i]*(1 - theta[i]) + theta[j]*(1 - theta[j]));
      K2 = ma*pow(theta[i] - theta[j], 2)/(ma + 1);
      d[target] = -0.5*K1*log(A) + K2*(1 - pow(A, 0.25*(ma + 1)));
      if (*variance)
	var[target] = pow(K1 + K2*0.5*(ma + 1)*pow(A, 0.25*(ma + 1)), 2)*Q[target]*(1 - Q[target])/(A*A*L[target]);
    }
  }
}

BF: base frequencies: a c g t
void dist_dna(char **x, int *n, int *s, int *model, double *d,
	      double *BF, int *pairdel, int *variance, double *var,
	      int *gamma, double *alpha)
{
  switch (*model) {
  case 1 : dist_dna_JC69(x, n, s, d, pairdel, variance,
			 var, gamma, alpha); break;
  case 2 : dist_dna_K80(x, n, s, d, pairdel, variance,
			var, gamma, alpha); break;
  case 3 : dist_dna_F81(x, n, s, d, BF, pairdel, variance,
			var, gamma, alpha); break;
  case 4 : dist_dna_K81(x, n, s, d, pairdel, variance,
			var); break;
  case 5 : dist_dna_F84(x, n, s, d, BF, pairdel, variance,
			var); break;
  case 6 : dist_dna_T92(x, n, s, d, BF, pairdel, variance,
			var); break;
  case 7 : dist_dna_TN93(x, n, s, d, BF, pairdel, variance,
			 var, gamma, alpha); break;
  case 8 : dist_dna_GG95(x, n, s, d, pairdel,
			 variance, var); break;
  }
}
"""
        

        



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: chain2psl.py 2899 2010-04-13 14:37:37Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
chain2psl.py - convert a chain file to a psl file
=================================================

:Author: Andreas Heger
:Release: $Id: chain2psl.py 2899 2010-04-13 14:37:37Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

convert a `chain <http://www.breyer.com/ucsc/htdocs/goldenPath/help/chain.html>`_ 
formatted file to a `psl <http://genome.ucsc.edu/FAQ/FAQformat.html#format2>`_
formatted file.

This tool is equivalent to chainToPsl except that it will not compute the number
of matching, mismatching, etc. bases and thus does not require the sequences.

The nomenclature the UCSC uses is for its chain files is :file:`targetToQuery.chain` for 
mapping ``query`` to ``target`` (reference). According to the UCSC documentation, 
``target`` is the first entry in ``chain`` files. 

I have been using the nomenclature ``QueryToTarget.psl``. In following this convention,
the correct way to converting a psl file is::

   python chain2psl.py < targetToQuery.chain > QueryToTarget.psl

If you would like to keep the TargetToQuery convention, you will need to add a pslSwap::

   python chain2psl.py < targetToQuery.chain | pslSwap stdin stdout > targetToQuery.psl

Usage
-----

Type::

   python chain2psl.py --help

for command line help.

Command line options
--------------------

""" 

import os
import sys
import re
import optparse

import CGAT.Experiment as E
import CGAT.Blat as Blat
import alignlib

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: chain2psl.py 2899 2010-04-13 14:37:37Z andreas $", 
                                    usage = globals()["__doc__"] )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    ## do sth
    ninput, nskipped, noutput = 0, 0, 0

    psl = None

    def chain_iterator( infile ):
        lines = []
        for line in options.stdin:
            
            if line.startswith("#"): continue
            if line.strip() == "": continue
            if line.startswith("chain"):
                if lines: yield lines
                lines = []
            lines.append( line )
            
        yield lines

    for lines in chain_iterator(options.stdin):
        
        ninput += 1
        psl = Blat.Match()

        ( _, 
          _, 
          psl.mSbjctId,
          target_length,
          target_strand,
          target_start,
          target_end,
          psl.mQueryId,
          query_length,
          query_strand,
          query_start, 
          query_end,
          alignment_id ) = lines[0][:-1].split()
        
        ( psl.mQueryStart, psl.mQueryEnd, psl.mQueryLength,
          psl.mSbjctStart, psl.mSbjctEnd, psl.mSbjctLength ) = \
        [ int(x) for x in 
          (query_start, 
           query_end,
           query_length,
           target_start, 
           target_end,
           target_length) ]

        map_query2target = alignlib.makeAlignmentBlocks()
        
        qstart, tstart = psl.mQueryStart, psl.mSbjctStart
        
        for line in lines[1:-1]:
            size, dt, dq = [int(x) for x in line[:-1].split() ]
            map_query2target.addDiagonal( qstart,
                                          qstart + size,
                                          tstart - qstart )
            qstart += size + dq
            tstart += size + dt

        size = int(lines[-1][:-1])

        map_query2target.addDiagonal( qstart,
                                      qstart + size,
                                      tstart - qstart )

        psl.fromMap( map_query2target )

        # sort out strand
        # target_strand is always positive
        assert( target_strand == "+" )

        # if query strand is negative
        if query_strand == "-": 
            # invert both query and target
            psl.switchTargetStrand()
            # manually invert the query coordinates
            psl.mQueryFrom, psl.mQueryTo = psl.mQueryLength - psl.mQueryTo, psl.mQueryLength - psl.mQueryFrom

        options.stdout.write("%s\n" % psl )
        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput,nskipped) )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
PamMatrices.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import string
import re

##-------------------------------------------------------------------------------------------------------------
## data section

# translate pam matrix
new_labels = "ACDEFGHIKLMNPQRSTVWY"
old_labels = "ARNDCQEGHILKMFPSTWYV"

# note: pam-matrices are in the literature cited as Mij, Mij gives the probability that j mutates into i.
# Therefore rows and columns are exchanged here.

pam1 = """
         Ala  Arg  Asn  Asp  Cys  Gln  Glu  Gly  His  Ile  Leu  Lys  Met  Phe  Pro  Ser  Thr  Trp  Tyr  Val
           A    R    N    D    C    Q    E    G    H    I    L    K    M    F    P    S    T    W    Y    V
Ala A  9867     2    9   10    3    8   17   21    2    6    4    2    6    2   22   35   32    0    2   18
Arg R     1  9913    1    0    1   10    0    0   10    3    1   19    4    1    4    6    1    8    0    1
Asn N     4     1 9822   36    0    4    6    6   21    3    1   13    0    1    2   20    9    1    4    1
Asp D     6     0   42 9859    0    6   53    6    4    1    0    3    0    0    1    5    3    0    0    1
Cys C     1     1    0    0 9973    0    0    0    1    1    0    0    0    0    1    5    1    0    3    2
Gln Q     3     9    4    5    0 9876   27    1   23    1    3    6    4    0    6    2    2    0    0    1
Glu E    10     0    7   56    0   35 9865    4    2    3    1    4    1    0    3    4    2    0    1    2
Gly G    21     1   12   11    1    3    7 9935    1    0    1    2    1    1    3   21    3    0    0    5
His H     1     8   18    3    1   20    1    0 9912    0    1    1    0    2    3    1    1    1    4    1
Ile I     2     2    3    1    2    1    2    0    0 9872    9    2   12    7    0    1    7    0    1   33
Leu L     3     1    3    0    0    6    1    1    4   22 9947    2   45   13    3    1    3    4    2   15
Lys K     2    37   25    6    0   12    7    2    2    4    1 9926   20    0    3    8   11    0    1    1
Met M     1     1    0    0    0    2    0    0    0    5    8    4 9874    1    0    1    2    0    0    4
Phe F     1     1    1    0    0    0    0    1    2    8    6    0    4 9946    0    2    1    3   28    0
Pro P    13     5    2    1    1    8    3    2    5    1    2    2    1    1 9926   12    4    0    0    2
Ser S    28    11   34    7   11    4    6   16    2    2    1    7    4    3   17 9840   38    5    2    2
Thr T    22     2   13    4    1    3    2    2    1   11    2    8    6    1    5   32 9871    0    2    9
Trp W     0     2    0    0    0    0    0    0    0    0    0    0    0    1    0    1    0 9976    1    0
Tyr Y     1     0    3    0    3    0    1    0    4    1    1    0    0   21    0    1    1    2 9945    1
Val V    13     2    1    1    3    2    2    3    3   57   11    1   17    1    3    2   10    0    2 9901
"""

pam250 = """
         Ala  Arg  Asn  Asp  Cys  Gln  Glu  Gly  His  Ile  Leu  Lys  Met  Phe  Pro  Ser  Thr  Trp  Tyr  Val
          A    R    N    D    C    Q    E    G    H    I    L    K    M    F    P    S    T    W    Y    V
Ala A    13    6    9    9    5    8    9   12    6    8    6    7    7    4   11   11   11    2    4    9
Arg R     3   17    4    3    2    5    3    2    6    3    2    9    4    1    4    4    3    7    2    2
Asn N     4    4    6    7    2    5    6    4    6    3    2    5    3    2    4    5    4    2    3    3
Asp D     5    4    8   11    1    7   10    5    6    3    2    5    3    1    4    5    5    1    2    3
Cys C     2    1    1    1   52    1    1    2    2    2    1    1    1    1    2    3    2    1    4    2
Gln Q     3    5    5    6    1   10    7    3    7    2    3    5    3    1    4    3    3    1    2    3
Glu E     5    4    7   11    1    9   12    5    6    3    2    5    3    1    4    5    5    1    2    3
Gly G    12    5   10   10    4    7    9   27    5    5    4    6    5    3    8   11    9    2    3    7
His H     2    5    5    4    2    7    4    2   15    2    2    3    2    2    3    3    2    2    3    2
Ile I     3    2    2    2    2    2    2    2    2   10    6    2    6    5    2    3    4    1    3    9
Leu L     6    4    4    3    2    6    4    3    5   15   34    4   20   13    5    4    6    6    7   13
Lys K     6   18   10    8    2   10    8    5    8    5    4   24    9    2    6    8    8    4    3    5
Met M     1    1    1    1    0    1    1    1    1    2    3    2    6    2    1    1    1    1    1    2
Phe F     2    1    2    1    1    1    1    1    3    5    6    1    4   32    1    2    2    4   20    3
Pro P     7    5    5    4    3    5    4    5    5    3    3    4    3    2   20    6    5    1    2    4
Ser S     9    6    8    7    7    6    7    9    6    5    4    7    5    3    9   10    9    4    4    6
Thr T     8    5    6    6    4    5    5    6    4    6    4    6    5    3    6    8   11    2    3    6
Trp W     0    2    0    0    0    0    0    0    1    0    1    0    0    1    0    1    0   55    1    0
Tyr Y     1    1    2    1    3    1    1    1    3    2    2    1    2   15    1    2    2    3   31    2
Val V     7    4    4    4    4    4    4    4    5    4   15   10    4   10    5    5    5   72    4   17
"""


##-------------------------------------------------------------------------------------------------------------
def ExponentiateMatrix( matrix ):

    l = len(matrix)
    new_matrix = []
    for row in range(0, l):
        new_matrix.append([0.0] * l)

        for col in range(0,l):
            v = 0
            for x in range(0,l):
                v = v + matrix[row][x] * matrix[col][x]
            new_matrix[row][col] = v
    return new_matrix

def NormalizeMatrix( matrix ):

    l = len(matrix)    
    # normalize matrix:
    for row in range(0,l):
        total = 0.0
        for col in range(0,l):
            total += matrix[row][col]
        for col in range(0,l):
            matrix[row][col] = float(matrix[row][col]) / total


def MultiplyMatrices( matrix_a, matrix_b ):

    l = len(matrix_b)
    m = len(matrix_a[0])
    new_matrix = []
    for row in range(0, l):
        new_matrix.append([0.0] * l)

        for col in range(0,m):
            v = 0
            for x in range(0,m):
                v = v + matrix_a[row][x] * matrix_b[x][col]
            new_matrix[row][col] = v
    return new_matrix

def PrintMatrix( matrix ):
    
    # print matrix
    for row in range(0,len(matrix)):
        print new_labels[row], string.join(map(lambda x: "%6.4f " % x,matrix[row]), "")


def CreateMatrix( matrix_string ):

    new_indices = {}
    for x in range(0,len(new_labels)):
        new_indices[new_labels[x]] = x

    # create matrix
    matrix = []
    for i in range(0,len(new_labels)):
        matrix.append([0.0] * len(new_labels))

    rx = re.compile("\s+") 

    row = 0
    for line in string.split(pam1, "\n")[3:]:

        data = map(string.atoi, rx.split(line)[2:])

        if len(data) <> len(old_labels):
            continue

        for col in range(0,len(old_labels)):
            #exchange row and col
            matrix[new_indices[old_labels[col]]][new_indices[old_labels[row]]] = data[col]

        row = row + 1    

    return matrix

def GetIdentityMatrix():

    matrix = []
    for i in range(0,len(new_labels)):
        matrix.append([0.0] * len(new_labels))

    for i in range(0,len(new_labels)):
        matrix[i][i] = 1.0
        
    return matrix

def GetMatrix( matrix_number ):

    current_matrix = CreateMatrix( pam1 )
    NormalizeMatrix( current_matrix )

    result_matrix = GetIdentityMatrix()

    x = matrix_number
    
    while x > 0:
        if x & 1:
            result_matrix = MultiplyMatrices(result_matrix, current_matrix)
        x = x >> 1
        current_matrix = ExponentiateMatrix(current_matrix)

    return result_matrix
##-------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    
    matrix = GetMatrix( 100 )
    PrintMatrix( matrix )










################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
graph2besthits.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python graph2besthits.py --help

Type::

   python graph2besthits.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import tempfile
import time
import optparse
import math

import CGAT.Experiment as E

USAGE="""python %s [OPTIONS] < graph.in > graph.out

Version: $Id: graph2besthits.py 2782 2009-09-10 11:40:29Z andreas $

Parse graph and only output best hits for each query per genome. Note that
the graph needs to be sorted by query_token and sbjct_token.

Options:
-h, --help                      print this message.
-v, --verbose=                  loglevel.
-p, --pattern-genome=           pattern to extract genome
-m, --method=                   method [blast|score|pid]
-s, --score-factor=             thresholding on score (multiplicative)
-i, --pide-factor=              thresholding on pide (additive)
""" % sys.argv[0]


param_score_threshold_factor = 1.0
param_pide_threshold_factor = 0.0

class Link:

    def __init__(self):
        self.mQueryToken, self.mSbjctToken, self.score = "","", 0
        
    def Read( self, line ):
        self.mQueryToken, self.mSbjctToken, self.score = line.split("\t")[:3]
        self.score = float(self.score)

    def __str__(self):
        return "\t".join( (self.mQueryToken, self.mSbjctToken, str(self.score)) )
    
##-------------------------------------------------------------------------------
def PrintMatches( matches,
                  options,
                  filter=None,
                  score_threshold_factor = 1.0 ):
    """print best match per organism (sorted by E-Value).

    If there are several matches with the same E-Value, all are printed.
    """

    orgs = matches.keys()
    orgs.sort()

    if options.method == "distance":
        f1 = lambda x,y: cmp(x.score, y.score)
        f2 = lambda x,y: x.score <= y
    elif options.method == "similarity":
        f1 = lambda x,y:  -cmp(x.score, y.score)
        f2 = lambda x,y: x.score >= y
        raise "unknown method"

    nbest = options.nbest
    
    for org in orgs:
        
        if filter and filter == org: continue

        m = matches[org]
        
        score_threshold = int(m[0].score * options.score_threshold_factor)

        if options.loglevel >= 2:
            print "# %i matches for %s, min_score=%i" % (len(m), org, score_threshold)

        if options.loglevel >= 3:
            for mm in m:
                print str(mm)
            
        m.sort( f1 )

        s = m[0].score
        x = 0

        if nbest:
            for x in range(min(nbest, len(m))):            
                print str(m[x])
        elif options.score_threshold_factor != 1.0:
            ## take best match and all within score threshold factor
            for x in range(len(m)):
                if f2( m[x], score_threshold):
                    print str(m[x])
        else:
            ## write all with same primary score
            while x < len(m) and s == m[x].score and f2( m[x], score_threshold) :            
                print str(m[x])
                x += 1

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: graph2besthits.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-p", "--pattern-genome", dest="pattern_genome", type="string",
                      help="genome_pattern."  )
    
    parser.add_option("-m", "--method", dest="method", type="string",
                      help="method to use."  )

    parser.add_option("-n", "--nbest", dest="nbest", type="int",
                      help="use nbest links."  )
    
    parser.add_option("-s", "--keep-self", dest="keep_self", action="store_true",
                      help="apply to links within same genome."  )

    parser.set_defaults(
        pattern_genome = "^([^|]+)|",
        method = "distance",
        keep_self = False,
        nbest = 0,
        score_threshold_factor = 1.0,
        )

    (options, args) = E.Start( parser )
    
    ninput, noutput, nskipped, nfailed = 0, 0, 0, 0

    last_query_token = None
    rx = re.compile(options.pattern_genome)

    matches = {}
    
    for line in sys.stdin:

        if line[0] == "#": continue
        
        link = Link()
        try:
            link.Read( line )
        except ValueError:
            sys.stderr.write( "parsing error in line %s\n" % line[:-1] )
            continue
        
        ninput += 1
        if link.mQueryToken != last_query_token :
            if last_query_token:
                if not options.keep_self:
                    filter = rx.search(last_query_token).groups()[0]
                else:
                    filter = None
                PrintMatches( matches,
                              options,
                              filter = filter )
                noutput += 1
            matches = {}
            last_query_token = link.mQueryToken

        org = rx.search(link.mSbjctToken).groups()[0]
        
        if org not in matches: matches[org] = []
        matches[org].append( link )

    if not options.keep_self:
        filter = rx.search(last_query_token).groups()[0]
    else:
        filter = None

    PrintMatches( matches,
                  options,
                  filter = filter )
    
    noutput += 1
    
    print "# ninput=%i, noutput=%i" % (\
        ninput, noutput )
    
    E.Stop()






################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
concatenate_sequences.py - concatenate sequences from multiple fasta files
==========================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads sequences from two or more :term:`fasta` formatted
files and outputs a new file with the sequences concatenated per
entry. 

All files must have the same number of sequences and the id of
the first file is output.

Usage
-----

Example::

   python concatenate_sequences.py a.fasta b.fasta > c.fasta

Type::

   python concatenate_sequences.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import re

USAGE="""python %s [OPTIONS] in1 in2 [...]



""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.FastaIterator as FastaIterator

##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: concatenate_sequences.py 2782 2009-09-10 11:40:29Z andreas $", usage = globals()["__doc__"])

    (options, args) = E.Start( parser )

    if len(args) < 2:
        raise ValueError("please supply at least two filenames to concatenate.")

    iterators = []
    for a in args:
        iterators.append(FastaIterator.FastaIterator( IOTools.openFile(a,"r") ) )

    ninput, noutput, nerrors = 0, 0, 0
    
    while 1:

        sequences = []
        ids = []

        for iterator in iterators:
            try:
                cur_record = iterator.next()
            except StopIteration:
                break

            sequences.append( re.sub( " ", "", cur_record.sequence) )
            ids.append( cur_record.title )

        if not sequences: break
        ninput += 1

        if len(sequences) != len(iterators):
            raise "unequal number of sequences in files."

        noutput += 1

        options.stdout.write( ">%s\n%s\n" % (ids[0],
                                             "".join( sequences )))

    E.info( "ninput=%i, noutput=%i, nerrors=%i" % (ninput, noutput, nerrors) )
    
    E.Stop()
    
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
contigs2random_sample.py
=============================================

:Author: Nick Ilott
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

When conducting a simulation study it is often useful to simulate contigs from the 
original genome/genomes from which your contigs have been derived. This can then 
provide an expectation dataset for downstream analyses.

Usage
-----

Example::

   python contigs2random_sample.py --help

Type::

   python contigs2random_sample_.py --help

for command line help.

Documentation
-------------

The script takes as input a multifasta set of contigs with arbitrary names as may have
been generated by any assembly algorithm. It also takes as input a map which maps the 
contig id to a genome id. The genomes that are referred to in the species-map file
are specified as a directory and these are where the samples are derived.

The contigs file is read from stdin and writes the result to stdout.

The species map MUST be in the format "contig" "species" and the species MUST be the 
contig identifier of the respective genome.

Code
----

'''

import os
import sys
import re
import optparse
import CGAT.Experiment as E
import CGAT.FastaIterator as FastaIterator
import collections
import itertools
import glob
import random
import CGAT.IOTools as IOTools

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser( version = "%prog version: $Id: contigs2random_sample.py 2871 2010-03-03 10:20:44Z nicki $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-m", "--species-map", dest="species_map", type="string",
                      help="text file specifying the mapping between contig and genome"  )

    parser.add_option("-g", "--genome-dir", dest="genome_dir", type="string",
                      help="specify directory where genome / genomes are stored"  )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    # read in contig lengths into dictionary
    E.info("reading contigs file")
    c_contigs = 0
    contigs_lengths = {}
    for fasta in FastaIterator.iterate(options.stdin):
        c_contigs += 1
        
        # titles of fasta records must be single strings with no special characters
        contigs_lengths[fasta.title.split(" ")[0]] = len(fasta.sequence)
        
    E.info("read %i contigs" % c_contigs)
    
    # read in mapping between spcies and contigs
    species_map = {}
    for line in open(options.species_map).readlines():
        data = line[:-1].split("\t")
        contig, species = data[0], data[1]
        species_map[contig] = species

    # read genomes into memory
    # NB this may need optimisin if using large 
    # genomes or many genomes
    E.info("reading genomes from %s" % options.genome_dir)

    # The directory must ONLY contain genome files!!
    genomes_sequences = {}
    c_genomes = 0
    for genome_file in glob.glob(os.path.join(options.genome_dir, "*")):
        c_genomes += 1
        for fasta in FastaIterator.iterate(IOTools.openFile(genome_file)):
            genomes_sequences[fasta.title] = fasta.sequence 
    E.info("read %i genomes from %s" % (c_genomes, options.genome_dir))
            
    # iterate over the contigs and sample from the respective genome
    E.info("iterating over contigs")
    c_contigs_output = 0
    for contig, length in contigs_lengths.iteritems():
        if contig not in species_map:
            E.warn("contig %s not in species map file" % contig)
        else:
            c_contigs_output += 1
            genome = species_map[contig]
            genome_length = len(genomes_sequences[genome])
        
            # get the start position from which to sample
            start = random.randint(1, genome_length)
            try:
                end = start + length - 1
            except ValueError:
                print "end of sampled contig extends beyond length of genome"
        
            sampled_seq = genomes_sequences[genome][start:end] 
            options.stdout.write( ">%s_random\n%s\n" % (contig + "_%s" % species_map[contig], sampled_seq))
        
    E.info("written %i contigs" % c_contigs_output)
    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



####
####
##
##
## Copyright (C) 2008 Andreas Heger All rights reserved
##
## Author: Andreas Heger <heger@ebi.ac.uk>
##
## $Id: bed2bed.py 2861 2010-02-23 17:36:32Z andreas $
##
##
####
####
'''
bed2bed.py - manipulate bed files
=================================

Purpose
-------

manipulate bed-formatted files.

The script currently implements the following methods:

1. merge: merge adjacent intervals. The option ``--merge-distance``
   permits the merging of segments that are up to a certain distance apart.

2. block: build blocked bed file (bed12 format) from individual blocks.

3. sanitize-genome: remove all empty intervals and intervals on unknown contigs. 
   Intervals extending beyond a contig a truncated. 
   
4. filter-genome: remove all intervals on unknown contigs or extending beyond 
   contigs.
5. extend: extend the start and end of each interval by a specified distance.
   Note: currently this method does not check that new interval is outside the confines of the contig

Usage
-----

Type::

   python bed2bed.py --help

for command line usage.

Methods
-------

This script provides several methods:

merge

filter-genome

bins
   merge adjacent bins by score. Several merging regimes are possible.
   equal-base, equal-intervals.
   This options requires the fifth field of the bed input file to be
   present.


Code
----
'''

import sys
import re
import string
import optparse
import time
import os
import itertools
import tempfile
import subprocess
import shutil

import CGAT.Experiment as E
import CGAT.Stats as Stats
import CGAT.GTF as GTF
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IOTools as IOTools
import CGAT.Bed as Bed
import pysam

def merge( iterator, max_distance = 0, by_name = False, min_intervals = 1 ):
    """iterator for merging adjacent bed entries.

    *max_distance* > 0 permits merging of intervals that are
    not directly adjacent.

    If *by_name = True*, only entries with the same name are merged.
    
    The score gives the number of intervals that have been merged.
    """

    def iterate_chunks( iterator ):
        last = iterator.next()
        max_end = last.end
        to_join = [last]

        for bed in iterator:
            d = bed.start - max_end
            if bed.contig == last.contig:
                assert bed.start >= last.start, "input file should be sorted by contig and position: d=%i:\n%s\n%s\n" % (d, last, bed)
            
            if bed.contig != last.contig or \
                    d > max_distance or \
                    (by_name and last.name != bed.name) :
                yield to_join
                to_join = []
                max_end = 0

            last = bed
            max_end = max( last.end, max_end )
            to_join.append( bed )

        if to_join: yield to_join
        raise StopIteration

    ninput, noutput, nskipped_min_intervals = 0, 0, 0

    for to_join in iterate_chunks(iterator):

        ninput += 1
        if len(to_join) < min_intervals: 
            nskipped_min_intervals += 1
            continue

        a = to_join[0]
        a.end = to_join[-1].end
        a.score = len(to_join)
        yield a
        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped_min_intervals=%i" % (ninput, noutput, nskipped_min_intervals) )

def filterGenome( iterator, contigs ):
    """remove bed intervals that are outside of contigs.
    
    contigs is a dictionary of contig sizes."""
    
    ninput, noutput = 0, 0
    nskipped_contig, nskipped_range, nskipped_endzero = 0, 0, 0

    for bed in iterator:
        ninput += 1
        if bed.contig not in contigs: 
            nskipped_contig += 1
            continue
        if bed.end >= contigs[bed.contig]:
            nskipped_range += 1
            continue
        if bed.end == 0:
            nskipped_endzero += 1
            continue    
        noutput += 1
        yield bed

    E.info( "ninput=%i, noutput=%i, nskipped_contig=%i, nskipped_range=%i, nskipped_endzero=%i" % \
                (ninput, noutput, nskipped_contig, nskipped_range, nskipped_endzero) )

def sanitizeGenome( iterator, contigs ):
    """truncate bed intervals that extend beyond contigs.

    removes empty intervals (start == end).

    throws an error if start > end.
    """
    
    ninput, noutput = 0, 0
    ntruncated_contig, nskipped_contig, nskipped_empty = 0, 0, 0

    for bed in iterator:
        ninput += 1
        if bed.contig not in contigs: 
            nskipped_contig += 1
            continue
        if bed.end >= contigs[bed.contig]:
            bed.end = contigs[bed.contig]
            ntruncated_contig += 1
        if bed.start < 0:
            bed.start = 0
            ntruncated_contig += 1
        if bed.start == bed.end:
            nskipped_empty += 1
            continue
        elif bed.start > bed.end:
            raise ValueError( "invalid interval: start > end for %s" % str(bed))

        noutput += 1
        yield bed

    E.info( "ninput=%i, noutput=%i, nskipped_contig=%i, ntruncated=%i, nskipped_empty=%i" % \
                (ninput, noutput, nskipped_contig, ntruncated_contig, nskipped_empty) )

def shiftIntervals( iterator, contigs, offset ):
    """shift intervals by a certain offset and ensure size is maintaned even id contig end reached.
    
    contigs is a dictionary of contig sizes."""
    
    ninput, noutput = 0, 0
    nskipped_contig, nskipped_range = 0, 0

    for bed in iterator:
        ninput += 1
        if bed.contig not in contigs: 
            nskipped_contig += 1
            continue
        if bed.end >= contigs[bed.contig]:
            nskipped_range += 1
            continue
        noutput += 1

        # add offset to each start and end, and adjust for contig length
        l = bed.end - bed.start
        newstart = bed.start + offset
        newend = bed.end + offset
        if newstart <0:
            newstart = 0
            newend = l
        if newend > contigs[bed.contig]:
            newstart = contigs[bed.contig] - l
            newend = contigs[bed.contig]

        bed.start = newstart
        bed.end = newend

        yield bed

    E.info( "ninput=%i, noutput=%i, nskipped_contig=%i, nskipped_range=%i" % \
                (ninput, noutput, nskipped_contig, nskipped_range) )

#IMS: new method: extend intervals by set amount
def extendInterval(iterator, distance):

    ninput, noutput, nskipped = 0, 0, 0
    for bed in iterator:
        ninput += 1
        newstart = bed.start - distance
        newend = bed.end + distance

        if newstart < 0:
            nskipped += 1
            continue

        bed.start = newstart
        bed.end = newend

        noutput += 1
        yield bed

    E.info("ninput = %i, noutput=%i, nskipped=%i" % (ninput,noutput,nskipped))


def main( argv = sys.argv ):

    parser = E.OptionParser( version = "%prog version: $Id: bed2bed.py 2861 2010-02-23 17:36:32Z andreas $", 
                                    usage = globals()["__doc__"] )

    #IMS: new method: extend intervals by set amount
    parser.add_option( "-m", "--method", dest="methods", type="choice", action="append",
                       choices=("merge", "filter-genome", "bins", "block", "sanitize-genome", "shift", "extend" ),
                       help="method to apply [default=%default]"  )

    parser.add_option( "--num-bins", dest="num_bins", type="int",
                      help="number of bins into which to merge (used for method `bins) [default=%default]"  )

    parser.add_option( "--bin-edges", dest="bin_edges", type="string",
                      help="bin_edges for binning method [default=%default]"  )

    parser.add_option( "--binning-method", dest="binning_method", type="choice",
                       choices = ("equal-bases", "equal-intervals", "equal-range" ),
                       help="method used for binning (used for method `bins` if no bin_edges is given) [default=%default]"  )

    parser.add_option( "--merge-distance", dest="merge_distance", type="int",
                      help="distance in bases over which to merge that are not directly adjacent [default=%default]"  )

    parser.add_option( "--merge-min-intervals", dest="merge_min_intervals", type="int",
                      help="only output merged intervals that are build from at least x intervals [default=%default]"  )

    parser.add_option( "--merge-by-name", dest="merge_by_name", action="store_true",
                      help="only merge intervals with the same name [default=%default]"  )

    parser.add_option( "--offset", dest="offset",  type="int",
                      help="offset for shifting intervals [default=%default]"  )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome."  )

    parser.add_option("-b", "--bam-file", dest="bam_file", type="string",
                      help="bam-formatted filename with genome."  )
    #IMS: new method: extend intervals by set amount
    parser.add_option("--extend-distance",dest="extend_distance", type="int",
                      help = "distance by which to extend intervals on either side for method=extend [default=%default]")

    parser.set_defaults( methods = [],
                         merge_distance = 0,
                         binning_method = "equal-bases",
                         genome_file = None,
                         bam_file = None,
                         num_bins = 5,
                         merge_min_intervals = 1,
                         bin_edges = None,
                         offset = 10000,
                         test = None,
                         extend_distance=1000)
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    contigs = None

    if options.genome_file:
        genome_fasta = IndexedFasta.IndexedFasta( options.genome_file )
        contigs = genome_fasta.getContigSizes()

    if options.bam_file:
        samfile = pysam.Samfile( options.bam_file )
        contigs = dict( zip(samfile.references, samfile.lengths) )

    processor = Bed.iterator( options.stdin )

    for method in options.methods:
        if method ==  "filter-genome":
            if not contigs: raise ValueError("please supply contig sizes" )
            processor = filterGenome( processor, contigs )
        elif method == "sanitize-genome":
            if not contigs: raise ValueError("please supply contig sizes" )
            processor = sanitizeGenome( processor, contigs )
        elif method == "merge":
            processor = merge( processor, options.merge_distance,
                               by_name = options.merge_by_name,
                               min_intervals = options.merge_min_intervals )
        elif method == "bins":
            if options.bin_edges: bin_edges = map(float, options.bin_edges.split(","))
            else: bin_edges = None
            processor, bin_edges = Bed.binIntervals( processor,
                                                     num_bins = options.num_bins,
                                                     method = options.binning_method,
                                                     bin_edges = bin_edges)
            E.info("# split bed: bin_edges=%s" % (str(bin_edges)))
            
        elif method == "block":
            processor = Bed.blocked_iterator( processor )
        elif method == "shift":
            processor = shiftIntervals( processor, contigs, offset=options.offset )
        #IMS: new method: extend intervals by set amount
        elif method == "extend":
            processor = extendInterval( processor, options.extend_distance )
            
    noutput = 0
    for bed in processor:
        options.stdout.write( str(bed) + "\n" )
        noutput += 1

    E.info( "noutput=%i" % (noutput) )

    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv ) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
psql_clone_database.py - clone a psql schema
============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script will clone a psql schema

Usage
-----

Example::

   python psql_clone_database.py src dest

will clone the schema ``src`` to ``dest``. 

Type::

   python psql_clone_database.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import time
import optparse
import math
import tempfile
import copy
import subprocess

import pgdb

import CGAT.Experiment as E

##---------------------------------------------------------------------------------------------
def DbExecute( dbhandle, statement ):

    cc = dbhandle.cursor()
    cc.execute(statement)
    result = cc.fetchall()
    cc.close()
    return result

##---------------------------------------------------------------------------------------------
def DbDo( dbhandle, statement ):

    cc = dbhandle.cursor()
    cc.execute(statement)
    dbhandle.commit()
    cc.close()

##------------------------------------------------------------------------------------------
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: psql_clone_database.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

    parser.set_defaults( \
        start = None,
        fields = [] )

    (options, args) = E.Start( parser, add_psql_options = True )
    
    dbhandle = pgdb.connect( options.psql_connection )

    if len(args) != 2:
        raise ValueError("please supply src and dest." )
    
    src, dest = args

    sql_get_schemas = "SELECT DISTINCT schemaname FROM pg_catalog.pg_stat_all_tables"
    sql_create_schema = "CREATE SCHEMA %s"

    cc = dbhandle.cursor()

    cc.execute( sql_get_schemas )
    databases = map(lambda x: x[0], cc.fetchall())
    cc.close()

    if src not in databases:
        raise ValueError("unknown schema %s" % src )

    if dest in databases:
        raise ValueError( "schema %s already exists" % (dest ) )

    host, database = options.psql_connection.split(":")

    cmd = 'pg_dump --ignore-version --schema-only --host=%(host)s --schema=%(src)s | sed "s/%(src)s/%(dest)s/g" | psql --host=%(host)s %(database)s' % (locals() ) 
    p = subprocess.Popen( cmd,
                          shell = True )
    sts = os.waitpid(p.pid, 0)

    cmd = 'pg_dump --ignore-version --data-only --host=%(host)s --schema=%(src)s | psql --host=%(host)s %(database)s' % (locals() ) 
    p = subprocess.Popen( cmd,
                          shell = True )
    sts = os.waitpid(p.pid, 0)

    E.Stop()

    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
cgat_log2wiki.py - convert a log file into a wiki page
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python cgat_log2wiki.py --help

Type::

   python cgat_log2wiki.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import CGAT.Experiment as E

if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: cgat_log2wiki.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option("-s", "--start", dest="start", type="string",
                      help="start of section." )

    parser.add_option("-e", "--end", dest="end", type="string",
                      help="end of section." )

    parser.add_option("-l", "--level", dest="level", type="int",
                      help="depth of sections." )

    parser.set_defaults(
        level = 2,
        start = None,
        end = None)

    (options, args) = E.Start( parser )

    if not options.start:
        keep = True
    else:
        keep = False

    section_id1 = 0
    section_id2 = 0
    last_l = None
    
    for line in sys.stdin:

        if options.start and not keep and re.search( options.start, line ):
            keep = True

        if not keep: continue
            
        if options.end and re.search( options.end, line ):
            break

        if line[0] != "#":
            l  = "| " + re.sub( "\t", " | ", line[:-1] ) + " | "
        else:
            
            if last_l:

                if re.match( "# ", line):
                    header = line[2:-1]
                    print last_l % header
                    continue
                
            last_l = None
                
            if re.match( "#----------------------------", line):
                section_id2 += 1
                last_l = "\n---%s subsection %i: %%s" % ("+" * (options.level+1), section_id2)
                l = None
            elif re.match( "#===========================", line):
                section_id1 += 1
                section_id2 = 0
                last_l = "\n---%s section %i: %%s" % ("+" * (options.level), section_id1)
                l = None
            else:
                l = line[:-1]

        if l:
            print l


    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
r_mann_whitney_u.py - Mann-Whitney U test
=========================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

Input: two sets of values. These can be either given
as values directly or as categories, which will be mapped to
values.

Usage
-----

Example::

   python r_mann_whitney_u.py --help

Type::

   python r_mann_whitney_u.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import time
import random
import rpy2
from rpy2.robjects import r as R
import CGAT.Experiment as E
import CGAT.IOTools as IOTools

if __name__  == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: r_mann_whitney_u.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option( "-m", "--method", dest="method", type="string",
                       help="method to use [ks=Kolmogorov-Smirnov,mwu=Mann-WhitneyU]")
    parser.add_option( "-a", "--hardcopy", dest="hardcopy", type="string",
                      help="write hardcopy to file.", metavar = "FILE" )
    parser.add_option( "-1", "--infile1", dest="filename_input1", type="string" ,
                       help="input filename for distribution 1.")
    parser.add_option( "-2", "--infile2", dest="filename_input2", type="string" ,
                       help="input filename for distribution 2.")
    parser.add_option( "-p", "--infile-map", dest="filename_input_map", type="string" ,
                       help="input filename for mapping categories to values.")

    parser.set_defaults(
        method = "ks",
        filename_input1 = None,
        filename_input2 = None,
        filename_input_map = None,
        )
    
    (options, args) = E.Start( parser,
                               add_pipe_options = True,
                               add_psql_options = True,)


    map_category2value = {}
    if options.filename_input_map:
        map_category2value = IOTools.ReadMap( open(options.filename_input_map, "r"),
                                              map_functions=(str,float))
    
    values1, errors1 = IOTools.ReadList( open(options.filename_input1, "r"),
                                         map_category=map_category2value )
    values2, errors2 = IOTools.ReadList( open(options.filename_input2, "r"),
                                         map_category=map_category2value )    
    
    E.info( "ninput1=%i, nerrors1=%i, ninput2=%i, nerrors2=%i" % (len(values1), len(errors1),
                                                                  len(values2), len(errors2)) )

    if options.hardcopy:
        R.png(options.hardcopy, width=1024, height=768)

    if options.method == "ks":
        result = R.ks_test( values1, values2 )
    elif options.method == "mwu":
        result = R.wilcox_test( values1, values2, paired=False)

    R.assign("v1", values1)
    R.assign("v2", values2)    

    R.layout(R.matrix((1,2,3,4), 2, 2, byrow = True))
        
    R.boxplot( values1, values2, col=('white','red'), main="Boxplot" )

    R("""qqplot( v1, v2, main ='Quantile-quantile plot' ); lines( c(0,1), c(0,1) );""")

    R("""hist( v1, freq=FALSE, width=0.5, density=10, main='Relative frequency histogram')""")
    R("""hist( v2, freq=FALSE, add=TRUE,   width=0.5, col='red', offset=0.5, density=20, angle=135)""")
    R("""hist( v1, freq=TRUE,  width=0.5, density=10, main='Absolute frequency histogram')""")
    R("""hist( v2, freq=TRUE,  add=TRUE,   width=0.5, col='red', offset=0.5, density=20, angle=135)""")

    print "## Results for %s" % result['method']
    for x in ['p.value', 'statistic', 'alternative', 'method']:
        print x, result[x]


    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mali_plain2aln.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python mali_plain2aln.py --help

Type::

   python mali_plain2aln.py --help

for command line help.

Documentation
-------------

Code
----

'''
import sys
import os
import string
import optparse

import CGAT.Experiment as E

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    # Length of each alignment segment
    LINE_WIDTH = 60

    # Get file data from stdin
    aliLines = sys.stdin.read().split("\n")
    while "" in aliLines: aliLines.remove("")
    while len(aliLines) > 2: del aliLines[2]

    # Sort data and get alignment length
    qArray = aliLines[0].split("\t")
    queryStart, queryAli, queryEnd, queryId = int(qArray[0]), qArray[1], int(qArray[2]), qArray[3]
    queryAli = queryAli.replace("-",".")
    sArray = aliLines[1].split("\t")
    sbjctStart, sbjctAli, sbjctEnd, sbjctId = int(sArray[0]), sArray[1], int(sArray[2]), sArray[3]
    sbjctAli = sbjctAli.replace("-",".")
    length = len(queryAli)

    CURR_POS = 0
    qFrom, qTo, sFrom, sTo = queryStart, 0, sbjctStart, 0
    while (CURR_POS < length):
	    if length - CURR_POS < 60: LINE_WIDTH = length - CURR_POS
	    qGaps, sGaps = 0, 0
	    # Get alignment slices
	    qAli = queryAli[CURR_POS:CURR_POS+LINE_WIDTH]
	    sAli = sbjctAli[CURR_POS:CURR_POS+LINE_WIDTH]
	    # Count gaps
	    for x in range(0,len(qAli),1):
		    if qAli[x] == ".": qGaps += 1
		    if sAli[x] == ".": sGaps += 1
	    qTo = qFrom + (LINE_WIDTH-1) - qGaps
	    sTo = sFrom + (LINE_WIDTH-1) - sGaps
	    print "\n"
	    print "%-20s %-10d%s\t%d" %(queryId, qFrom, qAli, qTo)
	    print "%-20s %-10d%s\t%d" %(sbjctId, sFrom, sAli, sTo)

	    qFrom = qTo+1
	    sFrom = sTo+1
	    CURR_POS += LINE_WIDTH

    print "\n"

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
mysql_clone_database.py - clone a mysql database
================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script clones a mysql database.

Usage
-----

Example::

   python mysql_clone_database.py src dest

will clone the database ``src`` to ``dest``. 

Type::

   python mysql_clone_database.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import CGAT.Experiment as E
import MySQLdb
import _mysql

USAGE="""python clone_database.py src dest

Clone a database by copying tables from one
database to another.
"""

def copyTables(dbhandle, dest, src, dry_run = False ):
    """copy tables from dest into src."""
    
    cc = dbhandle.cursor()
    cc.execute("SHOW tables FROM %s" % src)
    src_tables = map(lambda x: x[0], cc.fetchall())        
    cc.close()

    cc = dbhandle.cursor()
    cc.execute("SHOW tables FROM %s" % dest)
    dest_tables = set(map(lambda x: x[0], cc.fetchall()))
    cc.close()

    cc = dbhandle.cursor()
    cc.execute( "USE %s" % dest )
        
    for table in src_tables:

        if table in dest_tables:
            delete_statement = "DROP TABLE %s.%s" % (dest, table)
            if dry_run:
                print delete_statement
            else:
                cc.execute( delete_statement )
                
        cc.execute( "SHOW CREATE TABLE %s.%s" % (src, table) )
        create_statement = cc.fetchone()[1]
        copy_statement = "INSERT INTO %s.%s SELECT * FROM %s.%s" % (dest, table, src, table )
        if dry_run:
            print create_statement
            print copy_statement
        else:
            cc.execute( create_statement )
            cc.execute( copy_statement )
    
    cc.close()
    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: mysql_clone_database.py 2782 2009-09-10 11:40:29Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-c", "--create", dest="create", action="store_true",
                      help="create target database if it does not exist."  )

    parser.add_option("-d", "--dry-run", dest="dry_run", action="store_true",
                      help="dry run, do not execute any commands."  )


    parser.set_defaults (
        connection = "db",
        create = False,
        dry_run = False,
    )

    (options, args) = E.Start( parser, add_mysql_options = True )

    if len(args) != 2:
        print USAGE
        sys.exit(1)
        
    src, dest = args

    if src == dest:
        if options.loglevel >= 1:
            options.stdlog.write("# src and dest identical - no copying performed")            
        
        sys.exit(0)
        E.Stop()

    dbhandle = MySQLdb.Connect( host = options.host,
                                user = options.user,
                                passwd = options.password,
                                port = options.port )
                                               
    cc = dbhandle.cursor()

    cc.execute("SHOW databases")
    databases = map(lambda x: x[0], cc.fetchall())
    cc.close()
    
    databases = map(lambda x: x[0], cc.fetchall())

    if src not in databases:
        raise "unknown database %s" % src

    if dest not in databases:
        if options.create:
            if options.loglevel >= 1:
                options.stdlog.write("# creating database %s" % src)
                
            if not options.dry_run:
                cc = dbhandle.cursor()
                cc.execute("CREATE DATABASE %s" % dest)
                cc.close()

    copyTables( dbhandle, dest, src, options.dry_run )
        
    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
SequenceProperties.py - compute properties of sequences
=======================================================

Classes for extracting and reporting sequence properties on nucleotide sequences.

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import os, sys, string, re, tempfile, subprocess, optparse, math, hashlib, base64

import Genomics
import IOTools

import Bio.Alphabet.IUPAC

class SequenceProperties(object):

    mPseudoCounts = 0
    
    def __init__(self ):

        self.mLength = 0
        self.mNCodons = 0
        
    def addProperties( self, other ):
        """add properties."""

        self.mLength += other.mLength
        self.mNCodons += other.mNCodons

    def updateProperties( self ):
        pass
    
    def loadSequence( self, in_sequence ):
        """load sequence properties from a sequence."""

        sequence = re.sub( "[ -.]", "", in_sequence).upper()

        self.mLength = len(sequence)
        self.mNCodons = len(sequence) / 3
        
    def __str__(self):
        return "\t".join( self.getFields() )

    def getFields( self ):
        self.updateProperties()        
        return []
    
    def getHeaders( self ):
        return []

###########################################################################
class SequencePropertiesSequence(SequenceProperties):

    sequence = ""
    def __init__( self ):
        SequenceProperties.__init__(self)
        
    def addProperties( self, other ):
        """add properties."""
        SequenceProperties.addProperties(self,other)                

    def updateProperties( self ):
        SequenceProperties.updateProperties(self)        

    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence( self, sequence )
        self.sequence = sequence

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        return fields + [ self.sequence,]
    
    def getHeaders( self ):
        fields = SequenceProperties.getHeaders(self)
        return fields + ["sequence"]

###########################################################################
class SequencePropertiesLength(SequenceProperties):

    mPseudoCounts = 0
    
    def __init__( self ):
        
        SequenceProperties.__init__(self)
        
    def addProperties( self, other ):
        """add properties."""
        
        SequenceProperties.addProperties(self,other)                

    def updateProperties( self ):
        SequenceProperties.updateProperties(self)        
    
    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""
        
        SequenceProperties.loadSequence( self, sequence )

    def getFields(self):
        
        fields = SequenceProperties.getFields(self)
        return fields + ["%i" % self.mLength,
                         "%i" % self.mNCodons]
    
    def getHeaders( self ):
        
        fields = SequenceProperties.getHeaders(self)
        return fields + ["length", "ncodons"]

###########################################################################
class SequencePropertiesHid(SequenceProperties):

    def __init__( self ):
        SequenceProperties.__init__(self)
        self.mHid = ""

    def addProperties( self, other ):
        """add properties."""
        SequenceProperties.addProperties(self,other)                

    def updateProperties( self ):
        SequenceProperties.updateProperties(self)        
    
    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence( self, sequence )
        
        # do the encryption
        h = hashlib.md5(sequence).digest()
        # map to printable letters: hid has length 22, so the padded '=' are
        # truncated. You have to add them, if you ever want to decode,
        # but who would do such a thing :=)

        r = base64.encodestring(h)[0:22]
    
        # finally substitute some characters:
        # '/' for '_', so we have legal file names
        # '[' for '+' and ']' for '=' for internet-applications
    
        hid = string.replace(r  , '/', '_')
        hid = string.replace(hid, '+', '[')
        hid = string.replace(hid, '=', ']')
                                                
        self.mHid = hid

    def getFields(self):       
        fields = SequenceProperties.getFields(self)
        return fields + [ self.mHid, ]

    def getHeaders( self ):
        fields = SequenceProperties.getHeaders(self)
        return fields + ["hid",]

###########################################################################
class SequencePropertiesCounts(SequenceProperties):
    
    def __init__(self, alphabet ):
        
        SequenceProperties.__init__(self)
        self.mAlphabet = alphabet
        self.mCounts = {}
        self.mCountsOthers = 0

        for x in self.mAlphabet: self.mCounts[x] = 0
        
    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )
        for na, count in other.mCounts.items():
            self.mCounts[na] += count
        self.mCountsOthers += self.mCountsOthers

    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""

        SequenceProperties.loadSequence( self, sequence )        

        ## counts of amino acids
        self.mCounts = {}
        for x in self.mAlphabet: self.mCounts[x] = 0
        self.mCountsOthers = 0

        for na in sequence.upper():
            try:    
                self.mCounts[na] += 1
            except KeyError:
                self.mCountsOthers +=1 
        
    def getFields(self):

        fields = SequenceProperties.getFields(self)
        fields.append( "%i" % self.mCountsOthers )
        t = 0
        
        for x in self.mAlphabet:
            fields.append( "%i" % self.mCounts[x] )
            t += self.mCounts[x]

        if t == 0:
            for x in self.mAlphabet:
                fields.append( "na" )
        else:
            for x in self.mAlphabet:
                fields.append( "%f" % (float(self.mCounts[x]) / t))

        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        fields.append( "nUnk")
        fields.extend( ["n%s" % x for x in self.mAlphabet ] )
        fields.extend( ["p%s" % x for x in self.mAlphabet ] )
        return fields

###########################################################################
class SequencePropertiesNA(SequenceProperties):
    
    def __init__(self, reference_usage = [] ):
        
        SequenceProperties.__init__(self)

        self.mCountsGC = 0
        self.mCountsAT = 0
        self.mCountsOthers = 0
        
        ## counts of amino acids
        self.mCountsNA = {}
        self.mAlphabet = Bio.Alphabet.IUPAC.unambiguous_dna.letters + "N"
        for x in self.mAlphabet: self.mCountsNA[x] = 0
        
    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )

        self.mCountsGC += other.mCountsGC
        self.mCountsAT += other.mCountsAT
        self.mCountsOthers += other.mCountsOthers
        
        for na, count in other.mCountsNA.items():
            self.mCountsNA[na] += count

    def loadSequence( self, sequence ):
        
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence( self, sequence )        

        ## counts of amino acids
        self.mCountsNA = {}
        for x in self.mAlphabet: self.mCountsNA[x] = 0

        for na in sequence.upper():
            if na in ('G','C'):
                self.mCountsGC += 1
            elif na in ('A', 'T'):
                self.mCountsAT += 1
            try:    
                self.mCountsNA[na] += 1
            except KeyError:
                self.mCountsOthers +=1 
                    
    def getFields(self):

        fields = SequenceProperties.getFields(self)
        fields.append( "%i" % self.mCountsOthers )
        t = 0
        
        for x in self.mAlphabet:
            fields.append( "%i" % self.mCountsNA[x] )
            t += self.mCountsNA[x]

        fields.append( "%i" % self.mCountsGC )
        fields.append( "%i" % self.mCountsAT )        

        if t == 0:
            for x in self.mAlphabet:
                fields.append( "na" )
        else:
            for x in self.mAlphabet:
                fields.append( "%f" % (float(self.mCountsNA[x]) / t))

        t2 = float(self.mCountsGC + self.mCountsAT)
        if t2 == 0:
            fields.append( "na" )
            fields.append( "na" )
        else:
            fields.append( "%f" % (float(self.mCountsGC) / t2 ))
            fields.append( "%f" % (float(self.mCountsAT) / t2 ))
        
        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        fields.append( "nUnk")
        for x in self.mAlphabet:
            fields.append( "n%s" % x )
        fields.append("nGC")
        fields.append("nAT")        
        
        for x in self.mAlphabet:
            fields.append( "p%s" % x )
            
        fields.append("pGC")
        fields.append("pAT")        

        return fields


#######################################################################
class SequencePropertiesCpg(SequenceProperties):
    '''compute CpG frequencies as well as nucleotide frequencies.'''

    def __init__(self, reference_usage = [] ):
        
        SequenceProperties.__init__(self)

        self.mCountsCpG = 0
        self.mCountsGC = 0
        self.mCountsAT = 0
        self.mCountsOthers = 0
        self.mCountsC = 0
        self.mCountsG = 0
        self.mCountsA = 0
        self.mCountsT = 0
        self.mCountsN = 0
        
    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )

        self.mCountsCpG += other.mCountsCpG
        self.mCountsGC += other.mCountsGC
        self.mCountsAT += other.mCountsAT
        self.mCountsOthers += other.mCountsOthers
        self.mCountsC = other.mCountsC
        self.mCountsG = other.mCountsG
        self.mCountsA = other.mCountsA
        self.mCountsT = other.mCountsT
        self.mCountsN = other.mCountsN

    def loadSequence( self, sequence, next_char = None ):
        """load sequence properties from a sequence.

        If next_char is given and a 'G' and last char in sequence is 'C',
        an additional CpG will be counted in order to correctly account
        for CpG at sequence boundaries.
        """
        SequenceProperties.loadSequence( self, sequence )              

        # Single nucleotide counts
        lastc = False
        for na in sequence.upper():
            if na == 'C':
                self.mCountsC += 1
                lastc = True
            elif na =='G':
                if lastc:
                    self.mCountsCpG += 1
                self.mCountsG += 1
                lastc = False
            elif na =='A':
                self.mCountsA += 1
                lastc = False
            elif na =='T':
                self.mCountsT += 1
                lastc = False
            elif na =='N':
                self.mCountsN += 1
                lastc = False
            else:
                self.mCountsOthers +=1
                lastc = False

        if next_char:
            if lastc and next_char.upper() == "G":
                self.mCountsCpG += 1
               
    def getFields(self):

        fields = SequenceProperties.getFields(self)
        fields.append("%i" % self.mCountsC)
        fields.append("%i" % self.mCountsG)        
        fields.append("%i" % self.mCountsA) 
        fields.append("%i" % self.mCountsT) 
        fields.append("%i" % self.mCountsN) 
        fields.append("%i" % self.mCountsOthers)
        fields.append( "%i" % (self.mCountsC + self.mCountsG))
        fields.append( "%i" % (self.mCountsA + self.mCountsT))
        fields.append( "%i" % self.mCountsCpG )

        if self.mLength > 0:
            fields.append( "%f" % (float(self.mCountsC) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsG) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsA) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsT) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsN) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsOthers) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsC + self.mCountsG) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsA + self.mCountsT) / float(self.mLength) ))
            fields.append( "%f" % (float(self.mCountsCpG) / (float(self.mLength)/2.0) ))
            if (self.mCountsC*self.mCountsG)/self.mLength > 0:
                fields.append( "%f" % ( (float(self.mCountsCpG)) / ((float(self.mCountsC)*float(self.mCountsG))/float(self.mLength)) ) )
            else:
                fields.append( "%f" % 0.0 )
        else:
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )
            fields.append( "%f" % 0.0 )

        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        fields.append("nC")    
        fields.append("nG")    
        fields.append("nA")    
        fields.append("nT")
        fields.append("nN")    
        fields.append("nUnk")
        fields.append("nGC")
        fields.append("nAT")
        fields.append("nCpG")
        fields.append("pC")    
        fields.append("pG")    
        fields.append("pA")    
        fields.append("pT")
        fields.append("pN")    
        fields.append("pUnk")
        fields.append("pGC")
        fields.append("pAT")
        fields.append("pCpG")
        fields.append("CpG_ObsExp")

        return fields

###########################################################################
class SequencePropertiesDegeneracy (SequenceProperties):
    """count degeneracy.

    The degeneracies for amino acids are::

       2: MW are non-degenerate.
       9: EDKNQHCYF are 2-fold degenerate.
       1: I is 3-fold degenerate
       5: VGATP are 4-fold degenerate.
       3: RLS are 2-fold and four-fold degenerate.
          Depending on the first two codons, the codons are counted
          as two or four-fold degenerate codons. This is encoded
          in the file Genomics.py.

    Note that the number degenerate sites is computed across all codon 
    positions.
    """
    def __init__(self):
        
        SequenceProperties.__init__(self)

        self.mNGC = 0
        self.mNSites1D, self.mNSites2D, self.mNSites3D, self.mNSites4D = 0, 0, 0, 0
        self.mNGC3   = 0
        self.mN2DGC3 =  0
        self.mN3DGC3 = 0
        self.mN4DGC3 = 0

        self.mNStopCodons = 0
        
        ## setup counting arrays
        ## nucleotide counts for each position (is not a sum of the counts
        ## per degenerate site, as the codon might be intelligible, e.g. GNN).
        self.mCounts = [ {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 },
                         {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 },                         
                         {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 } ]

        ## nucleotide counts for each position per degeneracy
        self.mCountsDegeneracy = []
        
        for x in (0,1,2):
            xx = []
            for y in range(5):
                yy = {}            
                for z in Bio.Alphabet.IUPAC.extended_dna.letters: yy[z] = 0
                xx.append( yy ) 
            self.mCountsDegeneracy.append( xx )

    def addProperties( self, other ):
        
        SequenceProperties.addProperties( self, other )
        self.mNStopCodons += other.mNStopCodons

        for x in (0,1,2):
            for y in range(5):
                for z in Bio.Alphabet.IUPAC.extended_dna.letters:
                    self.mCountsDegeneracy[x][y][z] += other.mCountsDegeneracy[x][y][z]

        for x in (0,1,2):
            for y, count in other.mCounts[x].items():
                self.mCounts[x][y] += count
        
    def loadSequence( self, sequence ):
        
        """load sequence properties from a sequence."""

        SequenceProperties.loadSequence( self, sequence )        
        
        ## uppercase all letters
        sequence = sequence.upper()
        
        self.mNStopCodons = 0

        ## setup counting arrays
        ## nucleotide counts for each position (is not a sum of the counts
        ## per degenerate site, as the codon might be intelligible, e.g. GNN).
        self.mCounts = [ {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 },
                         {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 },                         
                         {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0 } ]

        ## nucleotide counts for each position per degeneracy
        self.mCountsDegeneracy = []
        
        for x in (0,1,2):
            xx = []
            for y in range(5):
                yy = {}            
                for z in Bio.Alphabet.IUPAC.extended_dna.letters: yy[z] = 0
                xx.append( yy ) 
            self.mCountsDegeneracy.append( xx )

        for codon in [ sequence[x:x+3] for x in range(0, len(sequence), 3)]:

            for x in (0,1,2): self.mCounts[x][codon[x]] += 1

            if Genomics.IsStopCodon( codon ):
                self.mNStopCodons += 1
                continue

            try:
                aa, deg1, deg2, deg3 = Genomics.GetDegeneracy( codon )
                degrees = (deg1, deg2, deg3)
                for x in range(len(degrees)):
                    self.mCountsDegeneracy[x][degrees[x]][codon[x]] += 1
                    
            except KeyError:
                pass

    def updateProperties( self ):
        """update fields from counts."""

        SequenceProperties.updateProperties( self )
        self.mNGC = 0
        self.mNSites1D, self.mNSites2D, self.mNSites3D, self.mNSites4D = 0, 0, 0, 0
        self.mNSitesD3 = 0
        
        for x in (0,1,2):
            self.mNGC += self.mCounts[x]['C'] + self.mCounts[x]['G']
            self.mNSites1D += sum(self.mCountsDegeneracy[x][1].values() )                                
            self.mNSites2D += sum(self.mCountsDegeneracy[x][2].values() )                    
            self.mNSites3D += sum(self.mCountsDegeneracy[x][3].values() )        
            self.mNSites4D += sum(self.mCountsDegeneracy[x][4].values() )        
            
        self.mNGC3  = self.mCounts[2]['C'] + self.mCounts[2]['G']
        self.mN2DGC3 = self.mCountsDegeneracy[2][2]['C'] + self.mCountsDegeneracy[2][2]['G']
        self.mN3DGC3 = self.mCountsDegeneracy[2][3]['C'] + self.mCountsDegeneracy[2][3]['G']
        self.mN4DGC3 = self.mCountsDegeneracy[2][4]['C'] + self.mCountsDegeneracy[2][4]['G']

        ## count all degenerate sites at third codon position
        self.mNSitesD3 = sum(self.mCountsDegeneracy[2][2].values() ) +\
                         sum(self.mCountsDegeneracy[2][3].values() ) +\
                         sum(self.mCountsDegeneracy[2][4].values() )    

        ## number of GCs at degenerate third codon positions
        self.mNDGC3 = self.mN2DGC3 + self.mN3DGC3 + self.mN4DGC3
        
        if self.mLength > 0:
            self.mPNGC = "%5.4f" % (float(self.mNGC) / float(self.mLength))
        else:
            self.mPNGC = "na"

        if self.mNCodons > 0:
            self.mPNGC3 = "%5.4f" % (float(self.mNGC3) / float(self.mNCodons))
        else:
            self.mPNGC3 = "na"

        if self.mNSites2D > 0:
            self.mPN2DGC3 = "%5.4f" % (float(self.mN2DGC3) / float(self.mNSites2D))
        else:
            self.mPN2DGC3 = "na"

        if self.mNSites3D > 0:
            self.mPN3DGC3 = "%5.4f" % (float(self.mN3DGC3) / float(self.mNSites3D))
        else:
            self.mPN3DGC3 = "na"

        if self.mNSites4D > 0:
            self.mPN4DGC3 = "%5.4f" % (float(self.mN4DGC3) / float(self.mNSites4D))
        else:
            self.mPN4DGC3 = "na"

        if self.mNSitesD3 > 0:
            self.mPNDGC3 = "%5.4f" % (float(self.mNDGC3) / float(self.mNSitesD3))
        else:
            self.mPNDGC3 = "na"

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        return fields + [ "%i" % self.mNStopCodons,
                          "%i" % self.mNSites1D,
                          "%i" % self.mNSites2D,
                          "%i" % self.mNSites3D,
                          "%i" % self.mNSites4D,
                          "%i" % self.mNSitesD3,
                          "%i" % self.mNGC,                           
                          "%i" % self.mNGC3,
                          "%i" % self.mNDGC3,                          
                          "%i" % self.mN2DGC3,                                                    
                          "%i" % self.mN3DGC3,                          
                          "%i" % self.mN4DGC3,
                          "%s" % self.mPNGC,
                          "%s" % self.mPNGC3,
                          "%s" % self.mPNDGC3,                                                                              
                          "%s" % self.mPN2DGC3,                                                    
                          "%s" % self.mPN3DGC3,                          
                          "%s" % self.mPN4DGC3 ]
    
    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        return fields + [  "nstops",
                           "nsites1d",
                           "nsites2d",
                           "nsites3d",                           
                           "nsites4d",
                           "nsitesd3",
                           "ngc",
                           "ngc3",
                           "ndgc3",
                           "n2dgc3", 
                           "n3dgc3",                          
                           "n4dgc3",
                           "pgc",
                           "pgc3",
                           "pdgc3", 
                           "p2dgc3",                           
                           "p3dgc3",                           
                           "p4dgc3",
                           ]
    

###########################################################################
class SequencePropertiesAA(SequenceProperties):
    
    mPseudoCounts = 0
    
    def __init__(self, reference_usage = [] ):
        
        SequenceProperties.__init__(self)
        
        self.mReferenceUsage = reference_usage

        ## counts of amino acids
        self.mCountsAA = {}

        for x in Bio.Alphabet.IUPAC.extended_protein.letters: self.mCountsAA[x] = 0
        
        self.mEntropy = 0
        
    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )

        for aa, count in other.mCountsAA.items():
            self.mCountsAA[aa] += count

    def loadSequence( self, sequence ):
        
        """load sequence properties from a sequence."""

        SequenceProperties.loadSequence( self, sequence )        

        ## counts of amino acids
        self.mCountsAA = {}

        for x in Bio.Alphabet.IUPAC.extended_protein.letters: self.mCountsAA[x] = 0
        
        for codon in [ sequence[x:x+3] for x in range(0, len(sequence), 3)]:
            aa = Genomics.MapCodon2AA( codon )
            self.mCountsAA[aa] += 1
                    
    def getFields(self):

        fields = SequenceProperties.getFields(self)

        t = 0
        
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "%i" % self.mCountsAA[x] )
            t += self.mCountsAA[x]
            
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "%f" % (float(self.mCountsAA[x]) / t))
        
        return fields

    def getHeaders( self ):
        
        fields = SequenceProperties.getHeaders(self)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "n%s" % x )
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "p%s" % x )

        return fields

###########################################################################
class SequencePropertiesEntropy(SequencePropertiesCounts):

    mPseudoCounts = 0
    
    def __init__(self, alphabet ):
        
        SequencePropertiesCounts.__init__(self, alphabet = alphabet )
        
        self.mEntropy = None

    def addProperties( self, other ):
        SequencePropertiesCounts.addProperties( self, other )
        
    def updateProperties( self ):

        SequencePropertiesCounts.updateProperties( self )
        
        self.mProperties = []

        self.mEntropy = self.getEntropy()
        
    def getEntropy( self, usage = None ):
        """return entropy of a source in terms of a reference usage.

        Also called conditional entropy or encoding cost.
        """

        e = None

        v = self.mCounts.values() 
        total = sum( v ) + len(v) * self.mPseudoCounts
        pc = self.mPseudoCounts 
        if total != 0:
            e = 0
            frequencies = [ float(x + pc) / total for x in v ]
            for f in frequencies:
                if f > 0: e -= f * math.log( f )
        return e

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        if self.mEntropy != None:
            fields.append("%f" % self.mEntropy)
        else:
            fields.append("na" )

        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        fields.append("entropy")
        return fields


###########################################################################
class SequencePropertiesCodons(SequenceProperties):

    mPseudoCounts = 0
    
    def __init__(self ):
        
        SequenceProperties.__init__(self)
        
        self.mCodonCounts = {}
        for c in Genomics.GeneticCodeAA.keys(): self.mCodonCounts[c] = 0

    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )
        
        for codon, count in other.mCodonCounts.items():
            if codon not in self.mCodonCounts: self.mCodonCounts[codon] = 0
            self.mCodonCounts[codon] += count

    def updateProperties( self ):

        SequenceProperties.updateProperties( self )
        
    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""

        if len(sequence) % 3:
            raise "sequence %s is not multiples of 3: length=%i!" % (cur_record.title, len(sequence))

        SequenceProperties.loadSequence( self, sequence )        

        ## uppercase all letters and count codons
        self.mCodonCounts = Genomics.CountCodons( sequence.upper() )

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        keys = self.mCodonCounts.keys()
        keys.sort()
        t = 0
        for key in keys:
            t += self.mCodonCounts[key]
            fields.append( "%i" % self.mCodonCounts[key] )
        t = float(t)
        for key in keys:
            fields.append( "%f" % ( float(self.mCodonCounts[key]) / t))
            
        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        
        keys = self.mCodonCounts.keys()
        keys.sort()
        for key in keys: fields.append( "n%s" % key )
        for key in keys: fields.append( "p%s" % key )        
            
        return fields

###########################################################################
class SequencePropertiesCodonUsage(SequencePropertiesCodons):

    mPseudoCounts = 0
    
    def __init__(self ):
        
        SequencePropertiesCodons.__init__(self)

        self.mCodonFrequencies = {}
        
        for codon, aa in Genomics.GeneticCodeAA.items():
            self.mCodonFrequencies[codon] = 0
        for codon in Genomics.StopCodons:
            self.mCodonFrequencies[codon] = 0
            
    def addProperties( self, other ):
        
        SequencePropertiesCodons.addProperties( self, other )

    def updateProperties( self ):

        SequencePropertiesCodons.updateProperties( self )

        self.mCodonFrequencies = Genomics.CalculateCodonFrequenciesFromCounts( self.mCodonCounts )
        
    def getFields(self):

        fields = SequenceProperties.getFields(self)

        keys = self.mCodonFrequencies.keys()
        keys.sort()
        for key in keys:
            fields.append( "%f" % self.mCodonFrequencies[key] )
            
        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        
        keys = self.mCodonFrequencies.keys()
        keys.sort()
        for key in keys: fields.append( "r%s" % key )        
            
        return fields

###########################################################################
class SequencePropertiesCodonTranslator(SequencePropertiesCodonUsage):
    """This class outputs the sequence with codons replaced by their frequencies.
    """
    mPseudoCounts = 0
    
    def __init__(self ):
        
        SequencePropertiesCodonUsage.__init__(self)
        
        self.mSequence = ""
        
    def addProperties( self, other ):

        self.mSequence += other.mSequence

    def loadSequence( self, sequence ):
        """load sequence properties from a sequence."""
        SequencePropertiesCodons.loadSequence( self, sequence )
        
        self.mSequence = sequence

    def getHeaders(self):

        fields = SequenceProperties.getHeaders(self)
        fields.append("tsequence")
        return fields

    def getFields( self ):

        fields = SequenceProperties.getFields(self)

        for x in range(0,len(self.mSequence), 3):
            codon = self.mSequence[x:x+3]
            if codon in self.mCodonFrequencies:
                v = self.mCodonFrequencies[codon]
            else:
                v = 0.0
                
            fields.append( "%i" % (v * 100) )
            
        return fields

###########################################################################
class SequencePropertiesBias(SequencePropertiesCodons):

    mPseudoCounts = 0
    
    def __init__(self, reference_usage = [] ):
        
        SequencePropertiesCodons.__init__(self)
        
        self.mReferenceUsage = reference_usage

        self.mEntropy = 0

    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )
        
    def updateProperties( self ):

        SequenceProperties.updateProperties( self )
        
        self.mProperties = []

        self.mEntropy = self.getEntropy()
        
        for usage in self.mReferenceUsage:
            self.mProperties.append( { 'ml': self.getMessageLength( usage ),
                                       'entropy': self.getEntropy( usage ),
                                       'kl': self.getKL( usage ) } )
        
    def getMessageLength( self, usage ):
        """return message length of a sequence in terms of its reference usage."""

        ml = 0

        for codon, count in self.mCodonCounts.items():
            ml -= float(count) * math.log( usage[codon] )

        return ml

    def getEntropy( self, usage = None ):
        """return entropy of a source in terms of a reference usage.
        Also called conditional entropy or encoding cost.

        Note that here I compute the sum over 20 entropies, one for each amino acid.

        If not given, calculate entropy.
        """

        e = 0

        freqs = Genomics.CalculateCodonFrequenciesFromCounts( self.mCodonCounts, self.mPseudoCounts )
        if usage == None: usage = freqs
        
        for codon, count in self.mCodonCounts.items():
            e -= freqs[codon] * math.log( usage[codon] )

        return e

    def getKL( self, usage ):
        """return Kullback-Leibler Divergence (relative entropy) of sequences with
        respect to reference codon usage.
        """

        e = 0

        freqs = Genomics.CalculateCodonFrequenciesFromCounts( self.mCodonCounts, self.mPseudoCounts )
        
        for codon, count in self.mCodonCounts.items():
            e += usage[codon] * math.log( usage[codon] / freqs[codon] )

        return e

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        fields.append("%f" % self.mEntropy)
        for x in self.mProperties:
            fields.append( "%f" % x['ml'] )
            fields.append( "%f" % (x['ml'] / self.mNCodons) )
            fields.append( "%f" % x['entropy'] )
            fields.append( "%f" % x['kl'] )

        return fields

    def getHeaders( self ):

        fields = SequenceProperties.getHeaders(self)
        
        fields.append("entropy")
        for x in range( len(self.mReferenceUsage) ):
            fields.append( "ml%i" % x )
            fields.append( "relml%i" % x )
            fields.append( "relentropy%i" % x )
            fields.append( "kl%i" % x )
            
        return fields

class SequencePropertiesAminoAcids(SequenceProperties):
    '''sequence properties of amino acid sequences.'''

    mPseudoCounts = 0

    def __init__(self, reference_usage = [] ):
        
        SequenceProperties.__init__(self)

        ## counts of amino acids
        self.mCountsAA = {}
        for x in Bio.Alphabet.IUPAC.extended_protein.letters: self.mCountsAA[x] = 0
        self.mOtherCounts = 0

    def addProperties( self, other ):
        SequenceProperties.addProperties( self, other )

        for aa, count in other.mCountsAA.items():
            self.mCountsAA[aa] += count

    def loadSequence( self, sequence ):
        """load sequence properties from a sequence.
        """

        SequenceProperties.loadSequence( self, sequence )        

        ## set to zero
        for x in Bio.Alphabet.IUPAC.extended_protein.letters: self.mCountsAA[x] = 0
        self.mOtherCounts = 0

        for aa in sequence: 
            if aa == "-": continue
            try:
                self.mCountsAA[aa] += 1
            except KeyError:
                self.mOtherCounts += 1

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        t = 0
        
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "%i" % self.mCountsAA[x] )
            t += self.mCountsAA[x]
            
        if t > 0:
            for x in Bio.Alphabet.IUPAC.extended_protein.letters:
                fields.append( "%f" % (float(self.mCountsAA[x]) / t))
        else:
            for x in Bio.Alphabet.IUPAC.extended_protein.letters:
                fields.append( "0" )

        return fields

    def getHeaders( self ):
        
        fields = SequenceProperties.getHeaders(self)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "n%s" % x )
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append( "p%s" % x )

        return fields




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
fasta2table.py - analyze sequences for codon bias and other sequence properties
====================================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script reads a collection of sequences in :term:`fasta` format and computes
various sequence properties (length, information content, codon bias, ...).

Usage
-----

Example::

   python cat in.fasta | fasta2table.py --sections=cpg

In this example we inout a fasta file and compute the sequence composition i.e.
%C, %G, %A, %T as well as dinucleotide (CpG) composition for each sequence in the 
set.


Type::

   python fasta2table.py --help

for command line help.

Documentation
-------------

This script was formerly called :file:`analyze_codonbias_shannon.py`.

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import math

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import CGAT.IOTools as IOTools
from CGAT.SequenceProperties import *
import CGAT.FastaIterator as FastaIterator

##---------------------------------------------------------------------------------------------
    
def main( argv = None ):

    parser = E.OptionParser( version = "%prog version: $Id: analyze_codonbias_shannon.py 2864 2010-03-03 10:18:16Z andreas $",
                                    usage = globals()["__doc__"] )

    parser.add_option("-w", "--filename-weights", dest="filename_weights", type="string",
                      help="filename with codon frequencies. Multiple filenames can be separated by comma." )
    
    parser.add_option("-s", "--section", dest="sections", type="string", action="append",
                      help="which sections to output. Possible choices are length|na|aa|cpg|degeneracy|bias|codons|codon-usage|codon-translator|sequence, [%default]" )

    parser.add_option("-t", "--type", dest="seqtype", type="choice",
                      choices=("na", "aa"),
                      help="type of sequence: na=nucleotides, aa=amino acids [%default].")

    parser.add_option("-e", "--regex-identifier", dest="regex_identifier", type="string",
                      help="regular expression to extract identifier from fasta description line." )

    parser.set_defaults(
        filename_weights = "uniform",
        pseudocounts = 1,
        sections = [],
        regex_identifier = "(.+)",
        seqtype = "na",
        )
    
    (options, args) = E.Start( parser, argv = argv )
    options.filename_weights = options.filename_weights.split(",")

    rx = re.compile( options.regex_identifier )

    reference_codons = []
    if options.filename_weights:
        for filename in options.filename_weights:
            if filename == "uniform":
                reference_codons.append( Genomics.GetUniformCodonUsage() )
            else:
                reference_codons.append( IOTools.ReadMap( open(filename, "r"), has_header = True, map_functions=(str, float) ) )

        ## print codon table differences
        options.stdlog.write("# Difference between supplied codon usage preferences.\n")
        for x in range(0, len(reference_codons)):
            for y in range(0, len(reference_codons)):
                if x == y: continue
                # calculate KL distance
                a = reference_codons[x]
                b = reference_codons[y]
                d = 0
                for codon, p in a.items():
                    if Genomics.IsStopCodon( codon ): continue
                    d += b[codon] * math.log( b[codon] / p )

                options.stdlog.write( "# tablediff\t%s\t%s\t%f\n" % (options.filename_weights[x],
                                                                     options.filename_weights[y],
                                                                     d ) )

    iterator = FastaIterator.FastaIterator( options.stdin )

    def getCounter( section ):

        if options.seqtype == "na":
            if section == "length":
                s = SequencePropertiesLength()
            elif section == "sequence":
                s = SequencePropertiesSequence()
            elif section == "hid":
                s = SequencePropertiesHid()
            elif section == "na":
                s = SequencePropertiesNA()
            elif section == "cpg":
                s = SequencePropertiesCpg()
            elif section == "aa":
                s = SequencePropertiesAA()
            elif section == "degeneracy":
                s = SequencePropertiesDegeneracy()
            elif section == "bias":
                s = SequencePropertiesBias( reference_codons )
            elif section == "codons":
                s = SequencePropertiesCodons()
            elif section == "codon-usage":
                s = SequencePropertiesCodonUsage()
            elif section == "codon-translator":
                s = SequencePropertiesCodonTranslator()
            else:
                raise ValueError("unknown section %s" % section)
        elif options.seqtype == "aa":
            if section == "length":
                s = SequencePropertiesLength()
            elif section == "sequence":
                s = SequencePropertiesSequence()
            elif section == "hid":
                s = SequencePropertiesHid()
            elif section == "aa":
                s = SequencePropertiesAminoAcids()
            else:
                raise ValueError("unknown section %s" % section)
        return s

    ## setup totals
    totals = {}
    for section in options.sections:
        totals[section] = getCounter( section )
        
    options.stdout.write ("id" )
    for section in options.sections:
        options.stdout.write( "\t" + "\t".join(totals[section].getHeaders()) )
        
    options.stdout.write("\n")
    options.stdout.flush()

    for cur_record in iterator:
        
        sequence = re.sub(" ", "", cur_record.sequence).upper()

        if len(sequence) == 0:
            raise ValueError( "empty sequence %s" % cur_record.title )

        id = rx.search( cur_record.title ).groups()[0]

        options.stdout.write("%s" % id)
        options.stdout.flush()

        for section in options.sections:
            s = getCounter( section )
            s.loadSequence( sequence )
            totals[section].addProperties( s )

            options.stdout.write( "\t" + "\t".join(s.getFields()) )
            
        options.stdout.write("\n")
        
    options.stdout.write( "total" )
    for section in options.sections:
        options.stdout.write( "\t" + "\t".join(totals[section].getFields()))
    options.stdout.write("\n")
    
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )
